/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#define PI 3.141592653

#include "FancyTasksApplet.h"
#include "FancyTasksConfiguration.h"
#include "FancyTasksIcon.h"
#include "FancyTasksItem.h"
#include "FancyTasksLauncher.h"
#include "FancyTasksSpacer.h"
#include "config.h"

#include <QDesktopWidget>
#include <QFontMetrics>
#include <QPainter>
#include <QSizeF>
#include <QTimer>

#ifdef FANCYTASKS_HAVE_COMPOSITING
#include <QX11Info>

#include <X11/extensions/Xrender.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/shape.h>
#include <X11/Xlib.h>
#include <fixx11h.h>
#endif

#include <Plasma/Theme>
#include <Plasma/Animator>
#include <Plasma/Containment>

#include <KRun>
#include <KMenu>
#include <KIcon>
#include <KLocale>
#include <KIconLoader>
#include <KWindowInfo>
#include <KConfigDialog>
#include <KWindowSystem>
#include <NETRootInfo>

#include <cmath>

K_EXPORT_PLASMA_APPLET(fancytasks, FancyTasksApplet)

FancyTasksApplet::FancyTasksApplet(QObject *parent, const QVariantList &args) : Plasma::Applet(parent, args)
{
    KGlobal::locale()->insertCatalog("fancytasks");

    setBackgroundHints(NoBackground);

    setHasConfigurationInterface(true);

    setAcceptDrops(true);

    KConfigGroup configuration = config();

    m_configuration = new FancyTasksConfiguration(qobject_cast<QObject*>(this));
    m_configuration->update(configuration);

    m_dropZone = new FancyTasksSpacer(this);

    m_layout = new QGraphicsLinearLayout(this);
    m_layout->setContentsMargins(2, 2, 2, 2);
    m_layout->setSpacing(0);
    m_layout->addItem(m_dropZone);

    m_dropZone->hide();

    m_theme = new Plasma::FrameSvg(this);
    m_theme->setImagePath("widgets/fancytasks");
    m_theme->setEnabledBorders(Plasma::FrameSvg::AllBorders);

    setLayout(m_layout);

    setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));

    setMaximumSize(INT_MAX, INT_MAX);

    setMinimumSize(0, 0);

    m_appletMaximumHeight = 100;

    m_initialFactor = 0.7;

    m_animation = 0;

    m_size = QSize(500, 100);

    resize(500, 100);
}

void FancyTasksApplet::init()
{
    KConfigGroup configuration = config();

    QStringList urls = configuration.readEntry("urls", QStringList());

    for (int i = 0; i < urls.size(); ++i)
    {
        addLauncher(launcherForUrl(urls.at(i)));
    }

    m_groupManager = new TaskManager::GroupManager(this);
    m_groupManager->setGroupingStrategy(m_configuration->groupingStrategy());
    m_groupManager->setSortingStrategy(m_configuration->sortingStrategy());
    m_groupManager->setShowOnlyCurrentDesktop(m_configuration->showOnlyCurrentDesktop());
    m_groupManager->setShowOnlyCurrentScreen(m_configuration->showOnlyCurrentScreen());
    m_groupManager->setShowOnlyMinimized(m_configuration->showOnlyMinimized());

    connect(this, SIGNAL(activate()), this, SLOT(showMenu()));
    connect(this, SIGNAL(sizeChanged(qreal)), m_dropZone, SLOT(setSize(qreal)));
    connect(m_groupManager->rootGroup(), SIGNAL(itemAdded(AbstractItemPtr)), this, SLOT(itemAdded(AbstractItemPtr)));
    connect(m_groupManager->rootGroup(), SIGNAL(itemRemoved(AbstractItemPtr)), this, SLOT(itemRemoved(AbstractItemPtr)));
    connect(m_groupManager->rootGroup(), SIGNAL(itemPositionChanged(AbstractItemPtr)), this, SLOT(itemPositionChanged(AbstractItemPtr)));

    if (containment())
    {
        m_groupManager->setScreen(containment()->screen());
    }

    m_groupManager->reconnect();

    if (!m_configuration->showOnlyLaunchers())
    {
        foreach (TaskManager::AbstractItemPtr abstractItem, m_groupManager->rootGroup()->members())
        {
            itemAdded(abstractItem);
        }
    }

    updateSize();
}

void FancyTasksApplet::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)

    painter->fillRect(rect(), Qt::transparent);

    if (formFactor() != Plasma::Horizontal && formFactor() != Plasma::Vertical)
    {
        painter->setRenderHint(QPainter::Antialiasing);
        painter->setCompositionMode(QPainter::CompositionMode_Source);

        m_theme->resizeFrame(m_size);
        m_theme->paintFrame(painter);
    }
}

void FancyTasksApplet::createConfigurationInterface(KConfigDialog *parent)
{
    KConfigGroup configuration = config();
    QWidget *generalWidget = new QWidget;
    QWidget *appearanceWidget = new QWidget;

    m_configuration->update(configuration);

    m_generalUi.setupUi(generalWidget);
    m_appearanceUi.setupUi(appearanceWidget);

    QStringList moveAnimationNames;
    moveAnimationNames.append(i18n("None"));
    moveAnimationNames.append(i18n("Zoom"));
    moveAnimationNames.append(i18n("Jump"));
    moveAnimationNames.append(i18n("Highlight"));
    moveAnimationNames.append(i18n("Spotlight"));
    moveAnimationNames.append(i18n("Fade"));

    QStringList moveAnimationIds;
    moveAnimationIds.append("none");
    moveAnimationIds.append("zoomAnimation");
    moveAnimationIds.append("jumpAnimation");
    moveAnimationIds.append("highlightAnimation");
    moveAnimationIds.append("spotlightAnimation");
    moveAnimationIds.append("fadeAnimation");

    for (int i = 0; i < moveAnimationIds.size(); ++i)
    {
        m_appearanceUi.moveAnimation->addItem(moveAnimationNames.at(i), moveAnimationIds.at(i));
    }

    QStringList iconAnimationNames;
    iconAnimationNames.append(i18n("None"));
    iconAnimationNames.append(i18n("Bounce"));
    iconAnimationNames.append(i18n("Zoom"));
    iconAnimationNames.append(i18n("Blink"));
    iconAnimationNames.append(i18n("Highlight"));
    iconAnimationNames.append(i18n("Spotlight"));
    iconAnimationNames.append(i18n("Rotate"));

    QStringList iconAnimationIds;
    iconAnimationIds.append("none");
    iconAnimationIds.append("bounceAnimation");
    iconAnimationIds.append("zoomAnimation");
    iconAnimationIds.append("blinkAnimation");
    iconAnimationIds.append("highlightAnimation");
    iconAnimationIds.append("spotlightAnimation");
    iconAnimationIds.append("rotateAnimation");

    for (int i = 0; i < iconAnimationIds.size(); ++i)
    {
        m_appearanceUi.demandsAttentionAnimation->addItem(iconAnimationNames.at(i), iconAnimationIds.at(i));
        m_appearanceUi.startupAnimation->addItem(iconAnimationNames.at(i), iconAnimationIds.at(i));
    }

    QStringList titleLabelModeNames;
    titleLabelModeNames.append(i18n("Never"));
    titleLabelModeNames.append(i18n("On mouse-over"));
    titleLabelModeNames.append(i18n("Always"));

    QStringList titleLabelModeIds;
    titleLabelModeIds.append("never");
    titleLabelModeIds.append("mouseOver");
    titleLabelModeIds.append("always");

    for (int i = 0; i < titleLabelModeIds.size(); ++i)
    {
        m_appearanceUi.titleLabelMode->addItem(titleLabelModeNames.at(i), titleLabelModeIds.at(i));
    }

    m_generalUi.groupingStrategy->addItem(i18n("Do Not Group"), QVariant(TaskManager::GroupManager::NoGrouping));
    m_generalUi.groupingStrategy->addItem(i18n("Manually"), QVariant(TaskManager::GroupManager::ManualGrouping));
    m_generalUi.groupingStrategy->addItem(i18n("By Program Name"), QVariant(TaskManager::GroupManager::ProgramGrouping));

    switch (m_groupManager->groupingStrategy())
    {
        case TaskManager::GroupManager::NoGrouping:
            m_generalUi.groupingStrategy->setCurrentIndex(0);
        break;
        case TaskManager::GroupManager::ManualGrouping:
            m_generalUi.groupingStrategy->setCurrentIndex(1);
        break;
        case TaskManager::GroupManager::ProgramGrouping:
            m_generalUi.groupingStrategy->setCurrentIndex(2);
        break;
        default:
            m_generalUi.groupingStrategy->setCurrentIndex(-1);
    }

    m_generalUi.sortingStrategy->addItem(i18n("Do Not Sort"), QVariant(TaskManager::GroupManager::NoSorting));
    m_generalUi.sortingStrategy->addItem(i18n("Manually"), QVariant(TaskManager::GroupManager::ManualSorting));
    m_generalUi.sortingStrategy->addItem(i18n("Alphabetically"), QVariant(TaskManager::GroupManager::AlphaSorting));
    m_generalUi.sortingStrategy->addItem(i18n("By Desktop"), QVariant(TaskManager::GroupManager::DesktopSorting));

    switch (m_groupManager->sortingStrategy())
    {
        case TaskManager::GroupManager::NoSorting:
            m_generalUi.sortingStrategy->setCurrentIndex(0);
        break;
        case TaskManager::GroupManager::ManualSorting:
            m_generalUi.sortingStrategy->setCurrentIndex(1);
        break;
        case TaskManager::GroupManager::AlphaSorting:
            m_generalUi.sortingStrategy->setCurrentIndex(2);
        break;
        case TaskManager::GroupManager::DesktopSorting:
            m_generalUi.sortingStrategy->setCurrentIndex(3);
        break;
        default:
            m_generalUi.sortingStrategy->setCurrentIndex(-1);
    }

    m_generalUi.showOnlyLaunchers->setChecked(m_configuration->showOnlyLaunchers());
    m_generalUi.showOnlyCurrentDesktop->setChecked(m_configuration->showOnlyCurrentDesktop());
    m_generalUi.showOnlyCurrentScreen->setChecked(m_configuration->showOnlyCurrentScreen());
    m_generalUi.showOnlyMinimized->setChecked(m_configuration->showOnlyMinimized());

    m_appearanceUi.drawReflections->setChecked(m_configuration->drawReflections());
    m_appearanceUi.useThumbnails->setChecked(m_configuration->useThumbnails());
    m_appearanceUi.titleLabelMode->setCurrentIndex(titleLabelModeIds.indexOf(m_configuration->titleLabelMode()));
    m_appearanceUi.moveAnimation->setCurrentIndex(moveAnimationIds.indexOf(m_configuration->moveAnimation()));
    m_appearanceUi.parabolicMoveAnimation->setChecked(m_configuration->parabolicMoveAnimation());
    m_appearanceUi.demandsAttentionAnimation->setCurrentIndex(iconAnimationIds.indexOf(m_configuration->demandsAttentionAnimation()));
    m_appearanceUi.startupAnimation->setCurrentIndex(iconAnimationIds.indexOf(m_configuration->startupAnimation()));
    m_appearanceUi.preallocateSpaceForIcons->setChecked(m_configuration->preallocateSpaceForIcons());

    moveAnimationTypeChanged(m_appearanceUi.moveAnimation->currentIndex());

    showOnlyLaunchersCheckedChanged(m_generalUi.showOnlyLaunchers->checkState());

    parent->addPage(generalWidget, i18n("General"), "go-home");
    parent->addPage(appearanceWidget, i18n("Appearance"), "preferences-desktop-theme");

    connect(parent, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
    connect(parent, SIGNAL(okClicked()), this, SLOT(configAccepted()));
    connect(m_generalUi.showOnlyLaunchers, SIGNAL(stateChanged(int)), this, SLOT(showOnlyLaunchersCheckedChanged(int)));
    connect(m_appearanceUi.moveAnimation, SIGNAL(currentIndexChanged(int)), this, SLOT(moveAnimationTypeChanged(int)));
}

void FancyTasksApplet::configAccepted()
{
    KConfigGroup configuration = config();

    if (!m_configuration->showOnlyLaunchers() && m_generalUi.showOnlyLaunchers->isChecked())
    {
        QHash<TaskManager::AbstractItemPtr, FancyTasksIcon*>::iterator iterator;

        for (iterator = m_taskIcons.begin(); iterator != m_taskIcons.end(); ++iterator)
        {
            layout()->removeItem(iterator.value());

            iterator.value()->deleteLater();
        }
    }
    else if (m_configuration->showOnlyLaunchers() && !m_generalUi.showOnlyLaunchers->isChecked())
    {
        foreach (TaskManager::AbstractItemPtr abstractItem, m_groupManager->rootGroup()->members())
        {
            itemAdded(abstractItem);
        }
    }

    configuration.writeEntry("moveAnimation", m_appearanceUi.moveAnimation->itemData(m_appearanceUi.moveAnimation->currentIndex()).toString());
    configuration.writeEntry("parabolicMoveAnimation", m_appearanceUi.parabolicMoveAnimation->isChecked());
    configuration.writeEntry("demandsAttentionAnimation", m_appearanceUi.demandsAttentionAnimation->itemData(m_appearanceUi.demandsAttentionAnimation->currentIndex()).toString());
    configuration.writeEntry("startupAnimation", m_appearanceUi.startupAnimation->itemData(m_appearanceUi.startupAnimation->currentIndex()).toString());
    configuration.writeEntry("preallocateSpaceForIcons", m_appearanceUi.preallocateSpaceForIcons->isChecked());
    configuration.writeEntry("drawReflections", m_appearanceUi.drawReflections->isChecked());
    configuration.writeEntry("useThumbnails",m_appearanceUi.useThumbnails->isChecked());
    configuration.writeEntry("titleLabelMode", m_appearanceUi.titleLabelMode->itemData(m_appearanceUi.titleLabelMode->currentIndex()).toString());
    configuration.writeEntry("showOnlyLaunchers", m_generalUi.showOnlyLaunchers->isChecked());
    configuration.writeEntry("showOnlyCurrentDesktop", m_generalUi.showOnlyCurrentDesktop->isChecked());
    configuration.writeEntry("showOnlyCurrentScreen", m_generalUi.showOnlyCurrentScreen->isChecked());
    configuration.writeEntry("showOnlyMinimized", m_generalUi.showOnlyMinimized->isChecked());
    configuration.writeEntry("groupingStrategy", m_generalUi.groupingStrategy->itemData(m_generalUi.groupingStrategy->currentIndex()).toInt());
    configuration.writeEntry("sortingStrategy", m_generalUi.sortingStrategy->itemData(m_generalUi.sortingStrategy->currentIndex()).toInt());

    m_configuration->update(configuration);

    m_groupManager->setGroupingStrategy(m_configuration->groupingStrategy());
    m_groupManager->setSortingStrategy(m_configuration->sortingStrategy());
    m_groupManager->setShowOnlyCurrentDesktop(m_configuration->showOnlyCurrentDesktop());
    m_groupManager->setShowOnlyCurrentScreen(m_configuration->showOnlyCurrentScreen());
    m_groupManager->setShowOnlyMinimized(m_configuration->showOnlyMinimized());

    if (containment())
    {
        m_groupManager->setScreen(containment()->screen());
    }

    m_groupManager->reconnect();

    for (int i = 0; i < layout()->count(); ++i)
    {
        if (layout()->itemAt(i) == m_dropZone)
        {
            continue;
        }

        FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(i);

        icon->reload();
    }

    updateSize();

    emit configNeedsSaving();
}

void FancyTasksApplet::saveConfig()
{
    KConfigGroup configuration = config();

    QStringList urls = configuration.readEntry("urls", QStringList());

    for (int i = 0; i < layout()->count(); ++i)
    {
        if (layout()->itemAt(i) == m_dropZone)
        {
            continue;
        }

        FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(i);

        if (icon->item()->itemType() == FancyTasksItem::Launcher)
        {
            urls.append(icon->item()->launcher()->url().url());
        }
    }

    configuration.writeEntry("urls", urls);

    emit configNeedsSaving();
}

void FancyTasksApplet::moveAnimationTypeChanged(int option)
{
    m_appearanceUi.parabolicMoveAnimation->setEnabled(option != 0 && m_appearanceUi.moveAnimation->itemData(option).toString() != "jumpAnimation");
    m_appearanceUi.preallocateSpaceForIcons->setEnabled(m_appearanceUi.moveAnimation->itemData(option).toString() == "zoomAnimation");
}

void FancyTasksApplet::showOnlyLaunchersCheckedChanged(int state)
{
    m_generalUi.showOnlyCurrentDesktop->setEnabled(!state);
    m_generalUi.showOnlyCurrentScreen->setEnabled(!state);
    m_generalUi.showOnlyMinimized->setEnabled(!state);
}

void FancyTasksApplet::constraintsEvent(Plasma::Constraints constraints)
{
    setBackgroundHints(NoBackground);

    if (constraints & Plasma::LocationConstraint)
    {
        layout()->setOrientation((formFactor() == Plasma::Vertical)?Qt::Vertical:Qt::Horizontal);
    }
}

void FancyTasksApplet::resizeEvent(QGraphicsSceneResizeEvent *event)
{
    if (formFactor() == Plasma::Vertical)
    {
        m_appletMaximumHeight = event->newSize().width();
    }
    else
    {
        m_appletMaximumHeight = event->newSize().height();
    }

    updateSize();
}

void FancyTasksApplet::dragMoveEvent(QGraphicsSceneDragDropEvent *event)
{
    int index = 0;

    if (formFactor() == Plasma::Vertical)
    {
        if (event->pos().y() > (boundingRect().height() / 2))
        {
            index = (layout()->count() - 1);
        }
    }
    else if (event->pos().x() > (boundingRect().width() / 2))
    {
        index = (layout()->count() - 1);
    }

    if (event->mimeData()->hasFormat("FancyTasksTask") || event->mimeData()->hasFormat("FancyTasksGroup"))
    {
        if (index < m_launcherIcons.count())
        {
            index = m_launcherIcons.count();
        }
    }
    else if (index > m_launcherIcons.count())
    {
        index = m_launcherIcons.count();
    }

    showDropZone(index);
}

void FancyTasksApplet::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
    Q_UNUSED(event)

    QTimer::singleShot(500, m_dropZone, SLOT(hide()));
}

void FancyTasksApplet::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    if (!KUrl::List::canDecode(event->mimeData()))
    {
        event->ignore();

        return;
    }

    KUrl::List droppedUrls = KUrl::List::fromMimeData(event->mimeData());

    for (int i = 0; i < droppedUrls.count(); ++i)
    {
        new KRun(droppedUrls[i], qobject_cast<QWidget*>(this));
    }
}

void FancyTasksApplet::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED(event)

    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation(5, 10, Plasma::Animator::LinearCurve, this, "moveAnimation");

    m_activeIcon = -1;
}

void FancyTasksApplet::wheelEvent(QGraphicsSceneWheelEvent *event)
{
    event->accept();

    int steps = (event->delta() / 120);

    m_activeIcon += steps;

    while (m_activeIcon < 0 || m_activeIcon >= (layout()->count() - 1))
    {
        if (m_activeIcon < 0)
        {
            m_activeIcon += (layout()->count() - 1);
        }
        else
        {
            m_activeIcon -= (layout()->count() - 1);
        }
    }

    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation(5, 10, Plasma::Animator::LinearCurve, this, "moveAnimation");

    if (event->modifiers() & Qt::ShiftModifier)
    {
        return;
    }

    FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(m_activeIcon + 1);
    icon->item()->activateWindow(true);
}

void FancyTasksApplet::itemAdded(AbstractItemPtr abstractItem)
{
    if (m_configuration->showOnlyLaunchers() || !abstractItem)
    {
        return;
    }

    FancyTasksIcon *icon = NULL;
    FancyTasksLauncher *launcher = NULL;

    if (abstractItem->isGroupItem() && m_groupManager->groupingStrategy() != TaskManager::GroupManager::ManualGrouping)
    {
        TaskManager::TaskGroup *group = static_cast<TaskManager::TaskGroup*>(abstractItem);

        if (group->name().isEmpty())
        {
            group->setName(static_cast<TaskManager::TaskItem*>(group->members().at(0))->task()->visibleName());
        }
    }

    QHash<FancyTasksIcon*, QDateTime>::iterator iterator;

    for (iterator = m_removedStartups.begin(); iterator != m_removedStartups.end(); ++iterator)
    {
        if (abstractItem->name().contains(iterator.key()->item()->title(), Qt::CaseInsensitive) && iterator.key()->item()->itemType() == FancyTasksItem::Startup)
        {
            icon = iterator.key();

            icon->item()->setTask(abstractItem);

            m_removedStartups.erase(iterator);

            break;
        }
    }

    for (int i = 0; i < m_launchers.count(); ++i)
    {
        if (!m_launchers.at(i)->isExecutable())
        {
            continue;
        }

        if (abstractItem->name().contains(m_launchers.at(i)->title(), Qt::CaseInsensitive))
        {
            launcher = m_launchers.at(i);
        }
    }

    if (!icon)
    {
        icon = new FancyTasksIcon(abstractItem, launcher, this);

        layout()->insertItem((m_groupManager->rootGroup()->members().indexOf(abstractItem) + m_launcherIcons.count() + (launcher?2:1)), icon);

        updateSize();
    }

    if (launcher && !icon->item()->launcher())
    {
        icon->item()->setLauncher(launcher);
    }

    m_taskIcons[abstractItem] = icon;
}

void FancyTasksApplet::itemRemoved(AbstractItemPtr abstractItem)
{
    if (m_configuration->showOnlyLaunchers() || !m_taskIcons.contains(abstractItem))
    {
        return;
    }

    FancyTasksIcon *icon = m_taskIcons[abstractItem];

    if (icon->item()->itemType() == FancyTasksItem::Startup)
    {
        m_removedStartups[icon] = QDateTime::currentDateTime();

        QTimer::singleShot(2000, this, SLOT(cleanupRemovedStartups()));

        return;
    }

    m_taskIcons.remove(abstractItem);

    if (!icon)
    {
       return;
    }

    layout()->removeItem(icon);

    icon->deleteLater();
}

void FancyTasksApplet::itemPositionChanged(AbstractItemPtr abstractItem)
{
    if (m_configuration->showOnlyLaunchers())
    {
        return;
    }

    FancyTasksIcon* icon = m_taskIcons[abstractItem];

    if (!icon)
    {
        return;
    }

    TaskManager::AbstractGroupableItem *temp = icon->item()->abstractItem();

    int index = (m_groupManager->rootGroup()->members().indexOf(temp) + m_launcherIcons.count() + 1);

    if ((layout()->itemAt(index)) == icon)
    {
        return;
    }

    layout()->removeItem(icon);
    layout()->insertItem(index, icon);
}

void FancyTasksApplet::addLauncher(FancyTasksLauncher *launcher, int index)
{
    if (!launcher || m_launcherIcons.contains(launcher))
    {
        return;
    }

    if (index < 0 || index > (m_launcherIcons.count() + 1))
    {
        index = (m_launcherIcons.count() + 1);
    }

    FancyTasksIcon *icon = new FancyTasksIcon(NULL, launcher, this);

    icon->setSize(m_iconSize);

    m_launcherIcons[launcher] = icon;

    layout()->insertItem(index, icon);

    updateSize();
}

void FancyTasksApplet::removeLauncher(FancyTasksLauncher *launcher)
{
    if (!m_launcherIcons.contains(launcher))
    {
        return;
    }

    if (launcher)
    {
        m_launchers.removeAll(launcher);
    }

    FancyTasksIcon *icon = m_launcherIcons[launcher];

    m_launcherIcons.remove(launcher);

    if (!icon)
    {
       return;
    }

    layout()->removeItem(icon);

    icon->deleteLater();
}

void FancyTasksApplet::removeLauncherIcon(FancyTasksLauncher *launcher)
{
    QHash<FancyTasksLauncher*, FancyTasksIcon*>::iterator iterator;

    for (iterator = m_launcherIcons.begin(); iterator != m_launcherIcons.end(); ++iterator)
    {
        if (iterator.value()->item()->launcher()->url() == launcher->url())
        {
            iterator.value()->deleteLater();

            m_launcherIcons.remove(launcher);

            break;
        }
    }
}

void FancyTasksApplet::cleanupRemovedStartups()
{
    QMutableHashIterator<FancyTasksIcon*, QDateTime> iterator(m_removedStartups);

    while (iterator.hasNext())
    {
        iterator.next();

        if (iterator.value().secsTo(QDateTime::currentDateTime()) > 1 && iterator.key()->item()->itemType() == FancyTasksItem::Startup)
        {
            iterator.key()->deleteLater();

            iterator.remove();
        }
    }
}

void FancyTasksApplet::iconHoverMoved(FancyTasksIcon *icon, qreal across)
{
    m_activeIcon = 0;

    if (icon)
    {
        for (int i = 0; i < layout()->count(); ++i)
        {
            if (layout()->itemAt(i) == m_dropZone)
            {
                continue;
            }

            FancyTasksIcon *tmp = (FancyTasksIcon*)layout()->itemAt(i);

            if (tmp == icon)
            {
                m_activeIcon = i;

                break;
            }
        }
    }

    m_across = across;

    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation(5, 10, Plasma::Animator::LinearCurve, this, "moveAnimation");
}

void FancyTasksApplet::iconHoverLeft()
{
    m_activeIcon = -1;

    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation(5, 10, Plasma::Animator::LinearCurve, this, "moveAnimation");
}

void FancyTasksApplet::moveAnimation(qreal progress)
{
    qreal factor;

    for (int i = 0; i < layout()->count(); ++i)
    {
        if (layout()->itemAt(i) == m_dropZone)
        {
            continue;
        }

        FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(i);

        if (i == m_activeIcon)
        {
            factor = 1;
        }
        else if (!m_configuration->parabolicMoveAnimation() || m_activeIcon < 0 || i < (m_activeIcon - 3) || i > (m_activeIcon + 3))
        {
            factor = m_initialFactor;
        }
        else if (i < m_activeIcon)
        {
            factor = (m_initialFactor + ((1 - m_initialFactor) * (cos(((i - m_activeIcon - m_across + 1) / 3) * PI) + 1)) / 2);
        }
        else
        {
            factor = (m_initialFactor + ((1 - m_initialFactor) * (cos(((i - m_activeIcon - m_across) / 3) * PI) + 1)) / 2);
        }

        if (icon->factor() != factor)
        {
            icon->setFactor(icon->factor() + ((factor - icon->factor()) * progress));
        }
    }

    updateMargins();
}

void FancyTasksApplet::showMenu()
{
    KMenu *menu = new KMenu;

    for (int i = 0; i < layout()->count(); ++i)
    {
        if (layout()->itemAt(i) == m_dropZone)
        {
            continue;
        }

        if (i == (m_launcherIcons.count() + 1))
        {
            menu->addSeparator();
        }

        FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(i);
        QAction *action = menu->addAction(icon->item()->icon(), icon->item()->title());
        QFont font = QFont(action->font());

        if (icon->item()->itemType() == FancyTasksItem::Launcher)
        {
            font.setItalic(true);
        }
        else if (icon->item()->demandsAttention())
        {
            font.setBold(true);
        }

        if (icon->item()->itemType() == FancyTasksItem::Group)
        {
            action->setMenu(new TaskManager::GroupPopupMenu(qobject_cast<QWidget*>(this), icon->item()->group(), m_groupManager));
        }
        else
        {
            connect(action, SIGNAL(triggered()), icon->item(), SLOT(activate()));
        }

        icon->item()->blockToolTip();

        action->setFont(font);
    }

    menu->exec(QCursor::pos());

    delete menu;
}

void FancyTasksApplet::showDropZone(int index)
{
    m_dropZone->setIndex(index);

    QTimer::singleShot(100, m_dropZone, SLOT(show()));
}

void FancyTasksApplet::updateSize()
{
    QSize size;

    m_iconSize = m_appletMaximumHeight;

    qreal iconSize = (m_iconSize * (m_configuration->drawReflections()?0.8:1));

    if (formFactor() != Plasma::Horizontal && formFactor() != Plasma::Vertical)
    {
        m_appletMaximumWidth = (m_theme->marginSize(Plasma::LeftMargin) + m_theme->marginSize(Plasma::RightMargin));
    }
    else
    {
        m_appletMaximumWidth = 0;
    }

    if (m_dropZone->isVisible())
    {
        m_appletMaximumWidth += iconSize;
    }

    for (int i = 0; i < (layout()->count() - 1); ++i)
    {
        if (m_configuration->moveAnimation() == "jumpAnimation" || (m_configuration->moveAnimation() == "zoomAnimation" && !m_configuration->preallocateSpaceForIcons() && (!i || (m_configuration->parabolicMoveAnimation() && i > 5))))
        {
            m_appletMaximumWidth += (m_initialFactor * iconSize);
        }
        else
        {
            m_appletMaximumWidth += iconSize;
        }
    }

    if (formFactor() == Plasma::Vertical)
    {
        size = QSize(m_appletMaximumHeight, m_appletMaximumWidth);

        setPreferredHeight(m_appletMaximumHeight);
    }
    else
    {
        size = QSize(m_appletMaximumWidth, m_appletMaximumHeight);

        setPreferredWidth(m_appletMaximumWidth);
    }

    if (size == m_size)
    {
        return;
    }

    m_size = size;

    emit sizeHintChanged(Qt::PreferredSize);

    resize(size);

    emit sizeChanged(m_iconSize);

    updateMargins();

    for (int i = 0; i < layout()->count(); ++i)
    {
        if (layout()->itemAt(i) == m_dropZone)
        {
            continue;
        }

        FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(i);
        icon->reload();
    }
}

void FancyTasksApplet::updateMargins()
{
    qreal width = 0;

    for (int i = 0; i < layout()->count(); ++i)
    {
        if (layout()->itemAt(i) == m_dropZone)
        {
            continue;
        }

        FancyTasksIcon *icon = (FancyTasksIcon*)layout()->itemAt(i);

        width += icon->boundingRect().width();
    }

    qreal margin = ((m_appletMaximumWidth - width) / 2);

    if (formFactor() == Plasma::Vertical)
    {
        layout()->setContentsMargins(0, margin, 0, margin);
    }
    else
    {
        layout()->setContentsMargins(margin, 0, margin, 0);
    }

    update();
}

void FancyTasksApplet::updateUrls(KUrl oldUrl, KUrl newUrl)
{
    KConfigGroup configuration = config();

    QStringList urls = configuration.readEntry("urls", QStringList());

    if (newUrl.url().isEmpty())
    {
        if (urls.indexOf(oldUrl.url()) > -1)
        {
            urls.removeAll(oldUrl.url());

            removeLauncher(launcherForUrl(oldUrl));
        }
    }
    else if (oldUrl.url().isEmpty())
    {
        if (urls.indexOf(newUrl.url()) == -1)
        {
            urls.append(newUrl.url());

            addLauncher(launcherForUrl(newUrl));
        }
    }
    else if (urls.indexOf(oldUrl.url()) > -1)
    {
        urls.replace(urls.indexOf(oldUrl.url()), newUrl.url());
    }

    configuration.writeEntry("urls", urls);

    emit configNeedsSaving();
}

FancyTasksLauncher* FancyTasksApplet::launcherForUrl(KUrl url)
{
    FancyTasksLauncher *launcher = NULL;

    if (!url.isValid())
    {
        return launcher;
    }

    for (int i = 0; i < m_launchers.count(); ++i)
    {
        if (m_launchers.at(i)->url() == url)
        {
            launcher = m_launchers.at(i);

            break;
        }
    }

    if (!launcher)
    {
        launcher = new FancyTasksLauncher(url, this);

        m_launchers.append(launcher);
    }

    return launcher;
}

FancyTasksConfiguration* FancyTasksApplet::configuration()
{
    return m_configuration;
}

FancyTasksSpacer* FancyTasksApplet::dropZone()
{
    return m_dropZone;
}

TaskManager::GroupManager* FancyTasksApplet::groupManager()
{
    return m_groupManager;
}

Plasma::FrameSvg* FancyTasksApplet::theme()
{
    return m_theme;
}

QGraphicsLinearLayout* FancyTasksApplet::layout()
{
    return m_layout;
}

QHash<FancyTasksLauncher*, FancyTasksIcon*> FancyTasksApplet::launcherIcons()
{
    return m_launcherIcons;
}

QHash<AbstractItemPtr, FancyTasksIcon*> FancyTasksApplet::taskIcons()
{
    return m_taskIcons;
}

QList<FancyTasksLauncher*> FancyTasksApplet::launchers()
{
    return m_launchers;
}

qreal FancyTasksApplet::initialFactor()
{
    return m_initialFactor;
}

qreal FancyTasksApplet::iconSize()
{
    return m_iconSize;
}

QPixmap FancyTasksApplet::windowPreview(WId window, int size)
{
    KWindowInfo *windowInformation = new KWindowInfo(window, NET::WMGeometry | NET::WMFrameExtents | NET::XAWMState);
    QPixmap thumbnail;

#ifdef FANCYTASKS_HAVE_COMPOSITING
    if (KWindowSystem::compositingActive())
    {
        Display *display = QX11Info::display();
        XWindowAttributes attributes;

        XCompositeRedirectWindow(display, window, CompositeRedirectAutomatic);

        XGetWindowAttributes(display, window, &attributes);

        XRenderPictFormat *format = XRenderFindVisualFormat(display, attributes.visual);

        if (format)
        {
            bool hasAlpha  = (format->type == PictTypeDirect && format->direct.alphaMask);
            int x = attributes.x;
            int y = attributes.y;
            int width = attributes.width;
            int height = attributes.height;

            XRenderPictureAttributes pictureAttributes;
            pictureAttributes.subwindow_mode = IncludeInferiors;

            Picture picture = XRenderCreatePicture(display, window, format, CPSubwindowMode, &pictureAttributes);

            XserverRegion region = XFixesCreateRegionFromWindow(display, window, WindowRegionBounding);

            XFixesTranslateRegion(display, region, -x, -y);
            XFixesSetPictureClipRegion(display, picture, 0, 0, region);
            XFixesDestroyRegion(display, region);

            XShapeSelectInput(display, window, ShapeNotifyMask);

            thumbnail = QPixmap(width, height);
            thumbnail.fill(Qt::transparent);

            XRenderComposite(display, (hasAlpha?PictOpOver:PictOpSrc), picture, None, thumbnail.x11PictureHandle(), 0, 0, 0, 0, 0, 0, width, height);
        }
    }
#endif

    if (thumbnail.isNull() && windowInformation->mappingState() == NET::Visible)
    {
        thumbnail = QPixmap::grabWindow(QApplication::desktop()->winId()).copy(windowInformation->frameGeometry());
    }

    if (!thumbnail.isNull())
    {
        if (thumbnail.width() > thumbnail.height())
        {
            thumbnail = thumbnail.scaledToWidth(size, Qt::SmoothTransformation);
        }
        else
        {
            thumbnail = thumbnail.scaledToHeight(size, Qt::SmoothTransformation);
        }
    }

    delete windowInformation;

    return thumbnail;
}

#include "FancyTasksApplet.moc"
