/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#define PI 3.141592653

#include "FancyTasksIcon.h"
#include "FancyTasksItem.h"
#include "FancyTasksApplet.h"
#include "FancyTasksConfiguration.h"
#include "FancyTasksLauncher.h"

#include <QPainter>
#include <QImage>
#include <QMimeData>
#include <QFontMetrics>
#include <QSizeF>
#include <QGraphicsView>
#include <QApplication>
#include <QTransform>

#include <Plasma/Svg>
#include <Plasma/Theme>
#include <Plasma/Corona>
#include <Plasma/Animator>
#include <Plasma/Containment>

#include <KMenu>
#include <KIcon>
#include <KLocale>
#include <KIconEffect>
#include <KIconLoader>
#include <KWindowInfo>
#include <KWindowSystem>
#include <NETRootInfo>

#include <cmath>

FancyTasksIcon::FancyTasksIcon(TaskManager::AbstractGroupableItem *abstractItem, FancyTasksLauncher *launcher, FancyTasksApplet *parent) : QGraphicsWidget(parent)
{
    m_applet = parent;

    m_animation = 0;

    m_animationRepeater = new QTimer(qobject_cast<QWidget*>(this));

    m_item =  new FancyTasksItem(abstractItem, launcher, this);

    m_layout = new QGraphicsLinearLayout(this);
    m_layout->addItem(m_item);

    setLayout(m_layout);

    m_pixmapValid = false;

    m_active = false;

    m_pressed = false;

    m_factor = m_applet->initialFactor();

    m_color = QColor();

    if (m_item->itemType() == FancyTasksItem::Group)
    {
        m_color = m_item->group()->color();
    }
    else if (m_item->itemType() == FancyTasksItem::Startup && m_applet->configuration()->startupAnimation() != "none")
    {
        startAnimation(1000, m_applet->configuration()->startupAnimation());
    }

    Plasma::Animator::self()->animateElement(this, Plasma::Animator::AppearAnimation);

    setSize(m_applet->iconSize());

    setPosition();

    connect(m_item, SIGNAL(pixmapChanged()), this, SLOT(refreshPixmap()));
    connect(m_item, SIGNAL(hoverMoved(FancyTasksIcon*, qreal)), m_applet, SLOT(iconHoverMoved(FancyTasksIcon*, qreal)));
    connect(m_item, SIGNAL(hoverLeft()), m_applet, SLOT(iconHoverLeft()));
    connect(m_applet, SIGNAL(sizeChanged(qreal)), this, SLOT(setSize(qreal)));
    connect(this, SIGNAL(destroyed()), m_applet, SLOT(updateSize()));
}

void FancyTasksIcon::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)

    QPixmap pixmap;
    qreal size = (m_size * 0.95 * ((m_applet->configuration()->moveAnimation() == "zoomAnimation")?m_factor:((m_applet->configuration()->moveAnimation() == "jumpAnimation")?m_applet->initialFactor():1)));

    if (m_pressed)
    {
        size *= 0.95;
    }

    if (size < 1)
    {
        return;
    }

    painter->setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    QPixmap target = QPixmap(boundingRect().size().toSize());
    target.fill(Qt::transparent);

    QPainter targetPainter(&target);
    targetPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    if ((m_animation || (m_applet->configuration()->moveAnimation() == "jumpAnimation" && m_progress < 1)) && !m_modifiedPixmap.isNull())
    {
        pixmap = m_modifiedPixmap;
    }
    else
    {
        pixmap = m_pixmap;
    }

    qreal xOffset = ((m_applet->formFactor() == Plasma::Vertical && m_applet->location() == Plasma::RightEdge)?(m_size - size):0);
    qreal yOffset = (((m_applet->formFactor() != Plasma::Vertical)?(m_size - size):0) - (m_size * 0.05));

    if (m_applet->configuration()->preallocateSpaceForIcons())
    {
        if (m_applet->formFactor() == Plasma::Vertical)
        {
            yOffset = ((m_size - size) / 2);
        }
        else
        {
            xOffset = ((m_size - size) / 2);
        }
    }

    if (m_applet->configuration()->moveAnimation() == "highlightAnimation" || m_applet->configuration()->moveAnimation() == "spotlightAnimation")
    {
        targetPainter.setOpacity(fabs((m_factor - m_applet->initialFactor()) * 2));
        targetPainter.drawPixmap(0, 0, ((m_applet->configuration()->moveAnimation() == "spotlightAnimation")?m_spotlightPixmap:m_highlightPixmap));
        targetPainter.setOpacity(0.8);
    }
    else if (m_applet->configuration()->moveAnimation() == "fadeAnimation")
    {
        targetPainter.setOpacity(1.5 - m_factor);
    }
    else if ((m_animation && m_currentAnimation == "jumpAnimation") || (m_active && m_progress < 1 && m_applet->configuration()->moveAnimation() == "jumpAnimation"))
    {
        yOffset *= ((cos(2 * PI * m_progress) + 1) / 4);
    }

    if (pixmap.isNull())
    {
        return;
    }

    if (m_active)
    {
        KIconEffect *effect = KIconLoader::global()->iconEffect();

        pixmap = effect->apply(pixmap, KIconLoader::Desktop, KIconLoader::ActiveState);
    }

    targetPainter.drawPixmap(QRectF(QPointF(xOffset, yOffset), QSizeF(size, size)), pixmap, QRectF(QPoint(0, 0), pixmap.size()));

    if ((m_applet->configuration()->titleLabelMode() == "always" || (m_applet->configuration()->titleLabelMode() == "mouseOver" && m_active)) && !m_item->title().isEmpty())
    {
        QString text = " " + m_item->title() + " ";

        QFont font = targetPainter.font();
        font.setPixelSize(size * 0.2);

        targetPainter.setFont(font);

        qreal textLength = (targetPainter.fontMetrics().width(m_item->title()) + (4 * targetPainter.fontMetrics().width(' ')));
        qreal textFieldWidth = (((textLength > size)?size:textLength) * 0.9);

        QRectF textField = QRectF(QPointF(((size - textFieldWidth) / 2), ((size + yOffset - (size * 0.3)))), QSizeF(textFieldWidth, (size * 0.25)));

        QPainterPath textFieldPath;
        textFieldPath.addRoundedRect(textField, 3, 3);

        targetPainter.setOpacity(0.6);
        targetPainter.fillPath(textFieldPath, QBrush(Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor)));
        targetPainter.setPen(QPen(Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor).darker()));
        targetPainter.drawRoundedRect(textField, 3, 3);
        targetPainter.setPen(QPen(Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor)));
        targetPainter.setOpacity(0.9);
        targetPainter.drawText(textField, (((textLength > textFieldWidth)?Qt::AlignLeft:Qt::AlignCenter) | Qt::TextIncludeTrailingSpaces), text);
    }

    if (m_applet->configuration()->drawReflections() && (m_applet->location() == Plasma::BottomEdge || m_applet->location() == Plasma::TopEdge || m_applet->location() == Plasma::Floating))
    {
        QPixmap reflection = target.copy(0, (target.height() * 0.5), target.width(), (target.height() * 0.25));
        reflection = reflection.transformed(QTransform(1, 0, 0, 0, -1, 0, 0, 0, 1), Qt::SmoothTransformation);

        QLinearGradient gradient(QPointF(0, 0), QPointF(0, reflection.height()));
        gradient.setColorAt(0, QColor(0, 0, 0, 150));
        gradient.setColorAt(0.3, QColor(0, 0, 0, 70));
        gradient.setColorAt(0.9, Qt::transparent);

        QPainter reflectionPainter(&reflection);
        reflectionPainter.setCompositionMode(QPainter::CompositionMode_DestinationIn);
        reflectionPainter.fillRect(0, 0, reflection.width(), reflection.height(), gradient);
        reflectionPainter.end();

        targetPainter.drawPixmap(QRectF(QPointF(0, ((yOffset + size) * 0.9)), reflection.size()), reflection, QRectF(QPoint(0, 0), reflection.size()));
    }

    if (m_item->itemsAmount())
    {
        int amount = m_item->itemsAmount();

        if (amount > 5)
        {
            amount = 5;
        }

        qreal diameter = (m_size * 0.15);
        qreal position = (((boundingRect().width() + xOffset) - (amount * diameter)) / 2);

        targetPainter.setOpacity(1);

        for (int i = 0; i < amount; ++i)
        {
            targetPainter.drawPixmap(position, (size + yOffset - diameter), m_taskMarkPixmap);

            position += diameter;
        }
    }

    painter->drawPixmap(0, 0, target);
}

void FancyTasksIcon::setFactor(qreal factor)
{
    if (factor == m_factor)
    {
        return;
    }

    m_factor = factor;

    if (!m_item)
    {
        return;
    }

    if (m_applet->configuration()->moveAnimation() == "zoomAnimation")
    {
        updateGeometry();

        setPosition();

        m_applet->layout()->invalidate();
    }

    update();

    if (!m_animation)
    {
        refreshPixmap(false);
    }
}

void FancyTasksIcon::setSize(qreal size)
{
    if (m_applet->configuration()->drawReflections())
    {
        size *= 0.8;
    }

    if (size == m_size)
    {
        return;
    }

    if (m_item->itemsAmount())
    {
        m_taskMarkPixmap = m_applet->theme()->pixmap("mark").scaled((size * 0.15), (size * 0.15), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    if (m_applet->configuration()->moveAnimation() == "highlightAnimation" || m_applet->configuration()->demandsAttentionAnimation() == "highlightAnimation" || m_applet->configuration()->startupAnimation() == "highlightAnimation")
    {
        m_highlightPixmap = m_applet->theme()->pixmap("highlight").scaled(size, size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    if (m_applet->configuration()->moveAnimation() == "spotlightAnimation" || m_applet->configuration()->demandsAttentionAnimation() == "spotlightAnimation" || m_applet->configuration()->startupAnimation() == "spotlightAnimation")
    {
        m_spotlightPixmap = m_applet->theme()->pixmap("spotlight").scaled(size, size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    if (m_color.isValid())
    {
        setColor(m_color);
    }

    if (size > m_size || m_pixmap.isNull())
    {
        m_pixmapValid = false;
    }

    m_size = size;

    if (!m_item)
    {
        return;
    }

    m_item->setIcon();

    m_item->setThumbnail();

    refreshPixmap(false);

    updateGeometry();

    setPosition();

    update();
}

void FancyTasksIcon::setActive(bool active)
{
    m_active = active;

    update();
}

void FancyTasksIcon::setPressed(bool pressed)
{
    m_pressed = pressed;

    update();
}

void FancyTasksIcon::setColor(QColor color)
{
    m_color = color;

    if (m_color.isValid())
    {
        KIconEffect effect;

        if (!m_taskMarkPixmap.isNull())
        {
            m_taskMarkPixmap = effect.apply(m_taskMarkPixmap, KIconEffect::Colorize, 1, m_color, true);
        }

        if (!m_highlightPixmap.isNull())
        {
            m_highlightPixmap = effect.apply(m_highlightPixmap, KIconEffect::Colorize, 1, m_color, true);
        }

        if (!m_spotlightPixmap.isNull())
        {
            m_spotlightPixmap = effect.apply(m_spotlightPixmap, KIconEffect::Colorize, 1, m_color, true);
        }
    }

    update();
}

void FancyTasksIcon::setPosition()
{
    qreal size = (m_size * ((m_applet->configuration()->moveAnimation() == "zoomAnimation")?m_factor:((m_applet->configuration()->moveAnimation() == "jumpAnimation")?m_applet->initialFactor():1)));

    if (m_applet->formFactor() == Plasma::Vertical)
    {
        if (m_applet->location() == Plasma::LeftEdge)
        {
            m_layout->setContentsMargins(0, 0, (m_size - size), 0);
        }
        else
        {
            m_layout->setContentsMargins(0, (m_size - size), 0, 0);
        }
    }
    else
    {
        m_layout->setContentsMargins(0, (m_size - size), 0, (boundingRect().height() - m_size));
    }
}

void FancyTasksIcon::refreshPixmap(bool force)
{
    if ((m_pixmapValid && !force) || m_size < 1)
    {
        return;
    }

    m_pixmap = QPixmap(m_size, m_size);
    m_pixmap.fill(Qt::transparent);

    QPainter pixmapPainter(&m_pixmap);
    pixmapPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    if (m_applet->configuration()->useThumbnails() && !m_item->thumbnail().isNull() && m_item->itemType() != FancyTasksItem::Group)
    {
        QPixmap pixmap;

        if (m_item->thumbnail().width() > m_item->thumbnail().height())
        {
            pixmap = m_item->thumbnail().scaledToWidth(m_size, Qt::SmoothTransformation);
        }
        else
        {
            pixmap = m_item->thumbnail().scaledToHeight(m_size, Qt::SmoothTransformation);
        }

        pixmapPainter.drawPixmap(((m_size - pixmap.width()) / 2), ((m_size - pixmap.height()) / 2), pixmap);
    }
    else
    {
        pixmapPainter.drawPixmap(0, 0, iconPixmap(m_size));
    }

    if (m_applet->configuration()->useThumbnails() && !m_item->thumbnail().isNull())
    {
        QPixmap icon = iconPixmap(m_size * 0.3);

        pixmapPainter.drawPixmap((m_pixmap.width() - icon.width()), (m_pixmap.height() - icon.height()), icon);
    }

    pixmapPainter.end();

    m_pixmapValid = true;

    if (m_item->itemType() != FancyTasksItem::Group)
    {
        m_item->updateToolTip();
    }

    update();
}

void FancyTasksIcon::startAnimation(int duration, QString animation)
{
    m_currentAnimationDuration = duration;

    m_currentAnimation = animation;

    m_animationRepeater->setInterval(duration);
    m_animationRepeater->start();

    repeatAnimation();

    connect(m_animationRepeater, SIGNAL(timeout()), this, SLOT(repeatAnimation()));
}

void FancyTasksIcon::stopAnimation()
{
    m_animationRepeater->stop();

    Plasma::Animator::self()->stopCustomAnimation(m_animation);

    m_animation = 0;
}

void FancyTasksIcon::repeatAnimation()
{
    if (m_animation)
    {
        Plasma::Animator::self()->stopCustomAnimation(m_animation);
    }

    m_animation = Plasma::Animator::self()->customAnimation((m_currentAnimationDuration / 100), m_currentAnimationDuration, Plasma::Animator::LinearCurve, this, ((m_currentAnimation == "spotlightAnimation")?"highlightAnimation":m_currentAnimation.toAscii()));
}

void FancyTasksIcon::bounceAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    int width = (m_size * ((progress < 0.5)?((progress * 0.5) + 0.5):(1 - (progress / 2))));

    int height = (m_size * ((progress < 0.5)?(1 - (progress / 2)):((progress * 0.5) + 0.5)));

    QPixmap scaled = m_pixmap.scaled(width, height, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (!scaled.isNull())
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.drawPixmap(((m_size - width) / 2), ((m_size - height) / 2), scaled);
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::jumpAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    int height = (m_size * ((progress > 0.8)?((progress > 0.95)?progress:(1.8 - progress)):1));

    QPixmap scaled = m_pixmap.scaled(m_size, height, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (!scaled.isNull())
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.drawPixmap(0, (m_size - height), scaled);
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    m_progress = progress;

    update();
}

void FancyTasksIcon::blinkAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    QPainter pixmapPainter(&pixmap);
    pixmapPainter.setOpacity(0.2 + ((cos(2 * PI * progress) + 0.5) / 4));
    pixmapPainter.drawPixmap(0, 0, m_pixmap);
    pixmapPainter.end();

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::highlightAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    QPixmap background = m_applet->theme()->pixmap((m_currentAnimation == "spotlightAnimation")?"spotlight":"highlight").scaled(m_size, m_size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (m_color.isValid())
    {
        KIconEffect effect;

        background = effect.apply(background, KIconEffect::Colorize, 1, m_color, true);
    }

    QPainter pixmapPainter(&pixmap);
    pixmapPainter.setOpacity((cos(2 * PI * progress) + 1) / 4);
    pixmapPainter.drawPixmap(0, 0, background);
    pixmapPainter.setOpacity(0.5);
    pixmapPainter.drawPixmap(0, 0, m_pixmap);
    pixmapPainter.end();

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::rotateAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    QPixmap scaled = m_pixmap.scaled((m_size * (0.5 + (((progress < 0.5)?progress:(1 - progress)) * 0.2))), (m_size * (0.5 + (((progress < 0.5)?progress:(1 - progress)) * 0.2))), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (!scaled.isNull())
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.translate((m_size / 2), (m_size / 2));
        pixmapPainter.rotate(360 * progress);
        pixmapPainter.drawPixmap(-(scaled.width() / 2), -(scaled.height() / 2), scaled);
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::zoomAnimation(qreal progress)
{
    QPixmap pixmap = QPixmap(m_size, m_size);
    pixmap.fill(Qt::transparent);

    qreal size = ((m_size * 0.75) + (m_size * ((cos(2 * PI * progress) + 0.5) / 8)));

    QPixmap scaled = m_pixmap.scaled(size, size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    if (!scaled.isNull())
    {
        QPainter pixmapPainter(&pixmap);
        pixmapPainter.drawPixmap(((m_size - size) / 2), ((m_size - size) / 2), scaled);
        pixmapPainter.end();
    }

    m_modifiedPixmap = pixmap;

    update();
}

void FancyTasksIcon::reload()
{
    if (m_item->itemType() == FancyTasksItem::Other)
    {
        return;
    }

    m_pixmapValid = false;

    prepareGeometryChange();

    update();
}

FancyTasksItem* FancyTasksIcon::item() const
{
    return m_item;
}

FancyTasksApplet* FancyTasksIcon::applet() const
{
    return m_applet;
}

qreal FancyTasksIcon::iconSize() const
{
     return m_size;
}

qreal FancyTasksIcon::factor() const
{
     return m_factor;
}

bool FancyTasksIcon::isActive() const
{
    return m_active;
}

QColor FancyTasksIcon::color() const
{
    return m_color;
}

QPixmap FancyTasksIcon::pixmap()
{
    return m_pixmap;
}

QPixmap FancyTasksIcon::iconPixmap(qreal size)
{
    QPixmap pixmap = m_item->icon().pixmap(size, size);

    if (pixmap.width() < size)
    {
        pixmap = pixmap.scaled(size, size, Qt::IgnoreAspectRatio, Qt::SmoothTransformation);
    }

    return pixmap;
}

QPoint FancyTasksIcon::popupPosition(QSize size, bool center)
{
    QRect geometry = iconGeometry();
    QPoint position(geometry.left(), geometry.top());

    switch (m_applet->location())
    {
        case Plasma::BottomEdge:
            position = QPoint((position.x() - (center?((size.width() - boundingRect().size().width()) / 2):0)), (position.y() - size.height()));
        break;
        case Plasma::TopEdge:
            position = QPoint((position.x() - (center?((size.width() - boundingRect().size().width()) / 2):0)), (position.y() + boundingRect().size().height()));
        break;
        case Plasma::LeftEdge:
            position = QPoint((position.x() + boundingRect().size().width()), (position.y() - (center?((size.height() - boundingRect().size().height()) / 2):0)));
        break;
        case Plasma::RightEdge:
            position = QPoint((position.x() - size.width()), (position.y() - (center?((size.height() - boundingRect().size().height()) / 2):0)));
        break;
        default:
            if (position.y() - size.height() > 0)
            {
                position = QPoint((position.x() - (center?((size.width() - boundingRect().size().width()) / 2):0)), (position.y() - size.height()));
            }
            else
            {
                position = QPoint((position.x() - (center?((size.width() - boundingRect().size().width()) / 2):0)), (position.y() + boundingRect().size().height()));
            }
    }

    QRect screenRect = m_applet->containment()->corona()->screenGeometry(m_applet->containment()?m_applet->containment()->screen():-1);

    if (m_applet->location() != Plasma::LeftEdge && position.rx() + size.width() > screenRect.right())
    {
        position.rx() -= ((position.rx() + size.width()) - screenRect.right());
    }

    if (m_applet->location() != Plasma::TopEdge && position.ry() + size.height() > screenRect.bottom())
    {
        position.ry() -= ((position.ry() + size.height()) - screenRect.bottom());
    }

    position.rx() = qMax(0, position.rx());

    return position;
}

QRect FancyTasksIcon::iconGeometry()
{
    qreal size = (m_size * 0.95 * ((m_applet->configuration()->moveAnimation() == "zoomAnimation")?m_factor:((m_applet->configuration()->moveAnimation() == "jumpAnimation")?m_applet->initialFactor():1)));

    QRect iconGeometry = QRect(QPoint(((m_applet->configuration()->preallocateSpaceForIcons() || m_applet->location() == Plasma::RightEdge)?(m_size - size):0), ((m_applet->configuration()->preallocateSpaceForIcons() || m_applet->location() == Plasma::BottomEdge || m_applet->location() == Plasma::Floating)?(m_size - size):0)), QSize(size, size));

    if (scene() && iconGeometry.isValid())
    {
        QGraphicsView *parentView = NULL;
        QGraphicsView *possibleParentView = NULL;

        foreach (QGraphicsView *view, scene()->views())
        {
            if (view->sceneRect().intersects(sceneBoundingRect()) || view->sceneRect().contains(scenePos()))
            {
                if (view->isActiveWindow())
                {
                    parentView = view;

                    break;
                }
                else
                {
                    possibleParentView = view;
                }
            }
        }

        if (!parentView)
        {
            parentView = possibleParentView;
        }

        if (parentView)
        {
            iconGeometry = parentView->mapFromScene(mapToScene(iconGeometry)).boundingRect().adjusted(0, 0, 1, 1);
            iconGeometry.moveTopLeft(parentView->mapToGlobal(iconGeometry.topLeft()));
        }
    }

    return iconGeometry;
}

QRectF FancyTasksIcon::boundingRect() const
{
    return QRectF(QPointF(0, 0), sizeHint(Qt::PreferredSize));
}

QSizeF FancyTasksIcon::sizeHint(Qt::SizeHint which, const QSizeF &constraint) const
{
    Q_UNUSED(which)
    Q_UNUSED(constraint)

    qreal size = (m_applet->configuration()->preallocateSpaceForIcons()?1:((m_applet->configuration()->moveAnimation() == "zoomAnimation")?m_factor:((m_applet->configuration()->moveAnimation() == "jumpAnimation")?m_applet->initialFactor():1)));
    qreal width = (m_size * ((m_applet->location() == Plasma::LeftEdge || m_applet->location() == Plasma::RightEdge)?(m_applet->configuration()->drawReflections()?1.4:1):size));
    qreal height = (m_size * ((m_applet->location() == Plasma::BottomEdge || m_applet->location() == Plasma::TopEdge || m_applet->location() == Plasma::Floating)?(m_applet->configuration()->drawReflections()?1.4:1):size));

    return QSizeF(width, height);
}
