/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "FancyTasksItem.h"
#include "FancyTasksIcon.h"
#include "FancyTasksApplet.h"
#include "FancyTasksConfiguration.h"
#include "FancyTasksLauncher.h"
#include "FancyTasksToolTip.h"
#include "FancyTasksSpacer.h"

#include <QSizeF>
#include <QPainter>
#include <QMimeData>
#include <QApplication>
#include <QFontMetrics>
#include <QGraphicsView>

#include <KMenu>
#include <KIcon>
#include <KLocale>

FancyTasksItem::FancyTasksItem(TaskManager::AbstractGroupableItem *abstractItem, FancyTasksLauncher *launcher, FancyTasksIcon *parent) : QGraphicsWidget(parent)
{
    setAcceptsHoverEvents(true);

    setAcceptDrops(true);

    m_abstractItem = abstractItem;

    m_launcher = launcher;

    m_parent = parent;

    m_itemsAmount = 0;

    m_demandsAttention = false;

    m_toolTipValid = false;

    m_toolTipTimer = new QTimer(qobject_cast<QWidget*>(this));

    m_dragTimer = new QTimer(qobject_cast<QWidget*>(this));

    m_toolTip = NULL;

    if (m_abstractItem)
    {
        setTask(m_abstractItem);
    }
    else if (m_launcher)
    {
        m_itemType = Launcher;
    }
    else
    {
        m_itemType = Other;
    }

    setLauncher(launcher);
}

void FancyTasksItem::hoverEnterEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED(event)

    m_parent->setActive(true);

    m_toolTipTimer->setInterval(1000);
    m_toolTipTimer->start();

    connect(m_toolTipTimer, SIGNAL(timeout()), this, SLOT(toggleToolTip()));

    if (m_parent->applet()->configuration()->moveAnimation() == "jumpAnimation")
    {
        Plasma::Animator::self()->customAnimation(50, 500, Plasma::Animator::LinearCurve, m_parent, "jumpAnimation");
    }
}

void FancyTasksItem::hoverMoveEvent(QGraphicsSceneHoverEvent *event)
{
    emit hoverMoved(m_parent, (qreal) (((m_parent->applet()->location() == Plasma::LeftEdge || m_parent->applet()->location() == Plasma::RightEdge)?event->pos().y():event->pos().x()) / m_parent->iconSize()));
}

void FancyTasksItem::hoverLeaveEvent(QGraphicsSceneHoverEvent *event)
{
    Q_UNUSED(event)

    m_parent->setActive(false);

    m_parent->setPressed(false);

    m_toolTipTimer->setInterval(200);
    m_toolTipTimer->start();

    connect(m_toolTipTimer, SIGNAL(timeout()), this, SLOT(toggleToolTip()));

    emit hoverLeft();
}

void FancyTasksItem::dragEnterEvent(QGraphicsSceneDragDropEvent *event)
{
    if (event->mimeData()->hasFormat("FancyTasksIcon"))
    {
        event->acceptProposedAction();

        return;
    }

    m_dragTimer->stop();

    if (m_itemType == Task || m_itemType == Group)
    {
        m_parent->setActive(true);

        m_dragTimer->setSingleShot(true);
        m_dragTimer->setInterval(300);
        m_dragTimer->start();

        connect(m_dragTimer, SIGNAL(timeout()), this, SLOT(activateWindow()));

        if (m_itemType == Group)
        {
            connect(m_dragTimer, SIGNAL(timeout()), this, SLOT(toggleToolTip()));
        }
    }

    if (m_itemType != Launcher)
    {
        event->ignore();
    }
}
void FancyTasksItem::dragMoveEvent(QGraphicsSceneDragDropEvent *event)
{
    if (((event->mimeData()->hasFormat("FancyTasksTask") || event->mimeData()->hasFormat("FancyTasksGroup")) && m_parent->applet()->groupManager()->sortingStrategy() == TaskManager::GroupManager::ManualSorting) || (m_parent->applet()->immutability() == Plasma::Mutable && (event->mimeData()->hasFormat("FancyTasksLauncher") || KUrl::List::canDecode(event->mimeData()))))
    {
        int index = 0;

        for (int i = 0; i < m_parent->applet()->layout()->count(); ++i)
        {
            if (m_parent->applet()->layout()->itemAt(i) == m_parent->applet()->dropZone())
            {
                continue;
            }

            FancyTasksIcon *icon = (FancyTasksIcon*)m_parent->applet()->layout()->itemAt(i);

            if (icon == m_parent)
            {
                break;
            }

            ++index;
        }

        if (m_parent->applet()->formFactor() == Plasma::Vertical)
        {
            if (event->pos().y() > (boundingRect().height() / 2))
            {
                ++index;
            }
        }
        else if (event->pos().x() > (boundingRect().width() / 2))
        {
            ++index;
        }

        if (event->mimeData()->hasFormat("FancyTasksTask") || event->mimeData()->hasFormat("FancyTasksGroup"))
        {
            if (index < m_parent->applet()->launcherIcons().count())
            {
                index = m_parent->applet()->launcherIcons().count();
            }
        }
        else if (index > m_parent->applet()->launcherIcons().count())
        {
            index = m_parent->applet()->launcherIcons().count();
        }

        m_parent->applet()->showDropZone(index);
    }
}

void FancyTasksItem::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
    QTimer::singleShot(500, m_parent->applet()->dropZone(), SLOT(hide()));

    if (m_itemType != Launcher)
    {
        event->ignore();
    }

    if (m_itemType == Group && m_toolTip && m_toolTip->isVisible())
    {
        toggleToolTip();
    }

    m_parent->setActive(false);

    m_dragTimer->stop();
}

void FancyTasksItem::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    m_parent->applet()->dropZone()->hide();

    if (event->mimeData()->hasFormat("FancyTasksIcon") && (m_itemType == Task || m_itemType == Group))
    {
        if (m_parent->applet()->groupManager()->groupingStrategy() == TaskManager::GroupManager::ManualGrouping)
        {
            QByteArray data = event->mimeData()->data("FancyTasksIcon");

            FancyTasksIcon *droppedIcon;

            memcpy(&droppedIcon, data, sizeof(FancyTasksIcon*));

            if (m_itemType == Group && droppedIcon->item()->itemType() == Task)
            {
                m_parent->applet()->groupManager()->manualGroupingRequest(droppedIcon->item()->abstractItem(), m_group);
            }
            else
            {
                TaskManager::ItemList items;
                TaskManager::ItemList temp;

                items.append(m_abstractItem);

                if (m_itemType == Group)
                {
                    temp = m_group->members();

                    for (int i = 0; i < temp.count(); ++i)
                    {
                        items.append(temp.at(i));
                    }
                }

                if (droppedIcon->item()->itemType() == Group)
                {
                    temp = droppedIcon->item()->group()->members();

                    for (int i = 0; i < temp.count(); ++i)
                    {
                        items.append(temp.at(i));
                    }
                }
                else
                {
                    items.append(droppedIcon->item()->abstractItem());
                }

                m_parent->applet()->groupManager()->manualGroupingRequest(items);
            }

            event->accept();
        }
        else
        {
            m_parent->applet()->dropEvent(event);

            event->ignore();
        }

        return;
    }

    if (m_itemType == Launcher)
    {
        m_launcher->dropEvent(m_parent, event);
    }

    event->ignore();
}

void FancyTasksItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    m_parent->setPressed(true);

    if (m_itemType != Task)
    {
        return;
    }

    m_task->task()->publishIconGeometry(m_parent->iconGeometry());

    event->accept();
}

void FancyTasksItem::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    m_parent->setPressed(false);

    if (QPoint(event->screenPos() - event->buttonDownScreenPos(Qt::LeftButton)).manhattanLength() < QApplication::startDragDistance() || m_itemType == Startup || (m_itemType == Launcher && m_parent->applet()->immutability() != Plasma::Mutable))
    {
        return;
    }

    blockToolTip();

    QByteArray data;
    data.resize(sizeof(FancyTasksIcon*));

    memcpy(data.data(), &m_parent, sizeof(FancyTasksIcon*));

    QMimeData* mimeData = new QMimeData();
    mimeData->setData("FancyTasksIcon", data);

    if (m_itemType == Launcher)
    {
        mimeData->setData("FancyTasksLauncher", QByteArray());

        m_launcher->url().populateMimeData(mimeData);
    }
    else if (m_itemType == Task)
    {
        mimeData->setData("FancyTasksTask", QByteArray());

        m_task->addMimeData(mimeData);
    }
    else if (m_itemType == Group)
    {
        mimeData->setData("FancyTasksGroup", QByteArray());

        m_group->addMimeData(mimeData);
    }
    else
    {
        mimeData->setData("FancyTasksOther", QByteArray());
    }

    QDrag *drag = new QDrag(event->widget());
    drag->setMimeData(mimeData);
    drag->setPixmap(m_parent->pixmap().scaled(32, 32, Qt::IgnoreAspectRatio, Qt::SmoothTransformation));
    drag->exec();
}

void FancyTasksItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    if (event->button() == Qt::LeftButton)
    {
        activate();

        event->accept();

        return;
    }
    else if (event->button() == Qt::MidButton)
    {
        if (m_itemType == Task)
        {
            m_task->task()->close();

            event->accept();

            return;
        }
        else if (m_itemType == Group)
        {
            m_group->close();

            event->accept();

            return;
        }
    }

    event->ignore();
}

void FancyTasksItem::contextMenuEvent(QGraphicsSceneContextMenuEvent *event)
{
    blockToolTip();

    if (m_itemType == Launcher)
    {
        KMenu *launcherMenu = m_launcher->contextMenu();

        launcherMenu->exec(m_parent->popupPosition(launcherMenu->sizeHint()));

        delete launcherMenu;
    }
    else if (m_itemType == Task || m_itemType == Group)
    {
        QList<QAction*> actionList;
        TaskManager::BasicMenu *taskMenu;

        if (m_itemType == Task)
        {
            taskMenu = new TaskManager::BasicMenu(qobject_cast<QWidget*>(this), m_task, m_parent->applet()->groupManager(), actionList);
        }
        else
        {
            taskMenu = new TaskManager::BasicMenu(qobject_cast<QWidget*>(this), m_group, m_parent->applet()->groupManager(), actionList);
        }

        if (m_launcher)
        {
            taskMenu->addSeparator();

            taskMenu->addMenu(m_launcher->contextMenu());
        }

        taskMenu->exec(m_parent->popupPosition(taskMenu->sizeHint()));

        delete taskMenu;

        event->accept();
    }
}

void FancyTasksItem::taskChanged(::TaskManager::TaskChanges changes)
{
    if (m_itemType != Task && m_itemType != Group)
    {
        return;
    }

    m_title = ((m_itemType == Group)?m_group->name():m_task->task()->visibleName());

    if (m_title.isEmpty())
    {
        if (m_itemType == Group)
        {
            m_title = static_cast<TaskManager::TaskItem*>(m_group->members().at(0))->task()->visibleName();

            m_group->setName(m_title);
        }
        else
        {
            m_title = i18n("Application");
        }
    }

    m_description = (m_abstractItem->isOnAllDesktops()?i18n("On all desktops"):i18nc("Which virtual desktop a window is currently on", "On %1", KWindowSystem::desktopName(m_abstractItem->desktop())));

    if (changes & TaskManager::IconChanged)
    {
        setIcon();
    }

    if (m_abstractItem->demandsAttention())
    {
        if (!m_demandsAttention)
        {
            m_parent->applet()->raise();
        }

        if (m_parent->applet()->configuration()->demandsAttentionAnimation() != "none")
        {
            m_parent->startAnimation(1000, m_parent->applet()->configuration()->demandsAttentionAnimation());

            m_demandsAttention = true;
        }
    }
    else if (m_demandsAttention)
    {
        m_parent->stopAnimation();

        m_demandsAttention = false;
    }

    if (changes & TaskManager::IconChanged || changes & TaskManager::NameChanged)
    {
        m_toolTipValid = false;

        if (m_toolTip && m_toolTip->isVisible())
        {
            updateToolTip();
        }
    }

    if (m_itemType == Task)
    {
        QTimer::singleShot(200, this, SLOT(setThumbnail()));
    }
}

void FancyTasksItem::launcherChanged()
{
    if (!m_launcher)
    {
        return;
    }

    setIcon();

    m_thumbnail = m_launcher->thumbnail();

    m_title = m_launcher->title();

    m_description = m_launcher->description();

    m_toolTipValid = false;

    emit pixmapChanged();
}

void FancyTasksItem::gotTaskPointer()
{
    setTask(m_abstractItem);
}

void FancyTasksItem::setIcon()
{
    switch (m_itemType)
    {
        case Launcher:
            m_icon = m_launcher->icon();
        break;
        case Startup:
            m_icon = KIcon(m_task->startup()->icon());
        break;
        case Task:
            m_icon = KIcon(m_task->task()->icon());
        break;
        case Group:
            m_icon = KIcon(m_group->icon());
        break;
        default:
            return;
    }

    emit pixmapChanged();
}

void FancyTasksItem::setThumbnail()
{
    if (m_itemType != Task)
    {
        return;
    }

    if (!m_parent->applet()->configuration()->useThumbnails() && !m_thumbnail.isNull())
    {
        emit pixmapChanged();

        return;
    }

    QPixmap thumbnail = FancyTasksApplet::windowPreview(m_task->task()->window(), ((200 > m_parent->iconSize())?200:m_parent->iconSize()));

    if (!thumbnail.isNull())
    {
        m_thumbnail = thumbnail;

        emit pixmapChanged();
    }
}

void FancyTasksItem::setItemType(ItemType itemType)
{
    m_itemType = itemType;
}

void FancyTasksItem::setLauncher(FancyTasksLauncher *launcher)
{
    if (m_launcher && m_itemType != Launcher)
    {
        m_launcher->removeItem();
    }

    m_launcher = launcher;

    if (m_launcher)
    {
        if (m_itemType != Launcher)
        {
            m_launcher->addItem(this);
        }

        launcherChanged();

        connect(m_launcher, SIGNAL(changed()), this, SLOT(launcherChanged()));
    }
}

void FancyTasksItem::setTask(TaskManager::AbstractGroupableItem *abstractItem)
{
    m_abstractItem = abstractItem;

    if (!m_abstractItem)
    {
        return;
    }

    if (m_abstractItem->isGroupItem())
    {
        m_group = static_cast<TaskManager::TaskGroup*>(m_abstractItem);

        m_itemType = Group;

        m_itemsAmount = m_group->members().count();

        connect(m_group, SIGNAL(changed(::TaskManager::TaskChanges)), this, SLOT(taskChanged(::TaskManager::TaskChanges)));
        connect(m_group, SIGNAL(groupEditRequest()), this, SLOT(groupDialog()));
        connect(m_group, SIGNAL(itemAdded(AbstractItemPtr)), this, SLOT(groupItemAdded(AbstractItemPtr)));
        connect(m_group, SIGNAL(itemRemoved(AbstractItemPtr)), this, SLOT(groupItemRemoved(AbstractItemPtr)));
    }
    else
    {
        m_task = static_cast<TaskManager::TaskItem*>(m_abstractItem);

        m_itemType = (m_task->task().isNull()?Startup:Task);

        m_itemsAmount = 1;

        connect(m_task, SIGNAL(changed(::TaskManager::TaskChanges)), this, SLOT(taskChanged(::TaskManager::TaskChanges)));
    }

    if (m_itemType == Startup)
    {
        m_title = m_task->startup()->text();

        m_description = i18n("Starting application...");

        connect(m_task, SIGNAL(gotTaskPointer()), this, SLOT(gotTaskPointer()));
    }
    else if (m_itemType == Task || m_itemType == Group)
    {
        m_parent->stopAnimation();

        taskChanged(::TaskManager::EverythingChanged);
    }
}

void FancyTasksItem::activate()
{
    switch (m_itemType)
    {
        case Launcher:
            m_launcher->activate();
        break;
        case Task:
            if (!m_task->task()->isActive() || m_task->task()->isIconified())
            {
                m_task->task()->activate();
            }
            else if (!m_task->task()->isActive() && !m_task->task()->isOnTop())
            {
                m_task->task()->raise();
            }
            else
            {
                m_task->task()->setIconified(true);
            }
        break;
        case Group:
            blockToolTip();

            if (m_toolTip && m_toolTip->isVisible())
            {
                m_toolTip->hide(true);
            }
            else
            {
                toggleToolTip();

                m_toolTip->setCanHide(false);
            }
        break;
        default:
        break;
    }
}

void FancyTasksItem::activateWindow(bool force)
{
    if (m_itemType == Task && (m_parent->isActive() || force))
    {
        m_task->task()->activateRaiseOrIconify();
    }
}

void FancyTasksItem::updateToolTip()
{
    if (m_toolTip)
    {
        if (m_toolTip->isVisible())
        {
            m_toolTip->hide(true);
        }

        delete m_toolTip;
    }

    if (m_itemType == Other)
    {
        return;
    }

    if (m_itemType == Task || m_itemType == Group)
    {
        m_toolTip = new FancyTasksToolTip(m_thumbnail, m_parent->iconPixmap(32), QString("<p><b>%1</b><br>\n<small>%2</small></p>").arg(m_title).arg(m_description), m_abstractItem, 200, qobject_cast<QWidget*>(this));
    }
    else
    {
        m_toolTip = new FancyTasksToolTip(m_thumbnail, m_parent->iconPixmap(32), QString("<p><b>%1</b><br>\n<small>%2</small></p>").arg(m_title).arg(m_description), m_launcher, 200, qobject_cast<QWidget*>(this));
    }

    m_toolTipValid = true;
}

void FancyTasksItem::updateToolTipPosition()
{
    m_toolTip->move(m_parent->popupPosition(m_toolTip->size(), true));
}

void FancyTasksItem::toggleToolTip()
{
    if (m_parent->isActive() && isVisible())
    {
        if (!m_toolTipValid)
        {
            updateToolTip();
        }

        updateToolTipPosition();

        m_toolTip->show();
    }
    else if (m_toolTip)
    {
        m_toolTip->hide();
    }

    m_toolTipTimer->stop();
}

void FancyTasksItem::blockToolTip()
{
    m_toolTipTimer->stop();

    if (m_toolTip && m_toolTip->isVisible())
    {
        m_toolTip->hide();
    }
}

void FancyTasksItem::groupDialogAccepted()
{
    m_group->setIcon(KIcon(m_groupUi.icon->icon()));
    m_group->setName(m_groupUi.name->text());
    m_group->setColor(m_groupUi.color->color());

    m_parent->setColor(m_groupUi.color->color());

    emit pixmapChanged();
}

void FancyTasksItem::groupDialogClosed()
{
    m_groupDialog = NULL;
}

void FancyTasksItem::groupDialog()
{
    if (m_itemType == Group && m_parent->applet()->groupManager()->taskGrouper()->editableGroupProperties() & TaskManager::AbstractGroupingStrategy::Name)
    {
        QWidget *groupWidget = new QWidget();

        m_groupUi.setupUi(groupWidget);
        m_groupUi.icon->setIcon(m_icon);
        m_groupUi.name->setText(m_title);
        m_groupUi.color->setColor(m_parent->color());

        m_groupDialog = new KDialog(qobject_cast<QWidget*>(this));
        m_groupDialog->setMainWidget(groupWidget);
        m_groupDialog->setButtons(KDialog::Cancel | KDialog::Ok);

        connect(m_groupDialog, SIGNAL(okClicked()), this, SLOT(groupDialogAccepted()));
        connect(m_groupDialog, SIGNAL(finished()), this, SLOT(groupDialogClosed()));

        m_groupDialog->setWindowTitle(i18n("%1 Settings", m_title));

        m_groupDialog->show();
    }
}

void FancyTasksItem::groupItemAdded(AbstractItemPtr item)
{
    Q_UNUSED(item)

    setTask(m_abstractItem);

    m_parent->update();
}

void FancyTasksItem::groupItemRemoved(AbstractItemPtr item)
{
    Q_UNUSED(item)

    setTask(m_abstractItem);

    m_parent->update();
}

KIcon FancyTasksItem::icon()
{
    return m_icon;
}

QPixmap FancyTasksItem::thumbnail()
{
    return m_thumbnail;
}

QString FancyTasksItem::title()
{
    return m_title;
}

QString FancyTasksItem::description()
{
    return m_description;
}

bool FancyTasksItem::demandsAttention()
{
    return m_demandsAttention;
}

int FancyTasksItem::itemsAmount()
{
    return m_itemsAmount;
}

FancyTasksItem::ItemType FancyTasksItem::itemType()
{
    return m_itemType;
}

TaskManager::AbstractGroupableItem* FancyTasksItem::abstractItem()
{
    return m_abstractItem;
}

TaskManager::TaskGroup* FancyTasksItem::group()
{
    return m_group;
}

FancyTasksLauncher* FancyTasksItem::launcher()
{
    return m_launcher;
}
