/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "FancyTasksLauncher.h"
#include "FancyTasksApplet.h"
#include "FancyTasksConfiguration.h"
#include "FancyTasksIcon.h"
#include "FancyTasksItem.h"

#include <QDir>
#include <QFileInfo>

#include <KShell>

FancyTasksLauncher::FancyTasksLauncher(KUrl url, FancyTasksApplet *parent)
{
    m_url = url;

    m_applet = parent;

    m_itemsAmount = 0;

    m_executable = false;

    setUrl(url);

    connect(this, SIGNAL(urlUpdated(KUrl, KUrl)), m_applet, SLOT(updateUrls(KUrl, KUrl)));
}

FancyTasksLauncher::~FancyTasksLauncher()
{
    for (int i = 0; i < m_items.count(); ++i)
    {
        m_items.at(i)->setLauncher(NULL);
    }
}

void FancyTasksLauncher::dropEvent(FancyTasksIcon *parent, QGraphicsSceneDragDropEvent *event)
{
    if (!KUrl::List::canDecode(event->mimeData()))
    {
        event->ignore();

        return;
    }

    KUrl::List urls = KUrl::List::fromMimeData(event->mimeData());

    if (!urls.count())
    {
        event->ignore();

        return;
    }

    if (m_mimetype->is("inode/directory"))
    {
        parent->item()->blockToolTip();

        Qt::DropAction dropAction = Qt::CopyAction;

        if ((event->modifiers() & Qt::ShiftModifier && event->modifiers() & Qt::ControlModifier) || event->modifiers() & Qt::AltModifier)
        {
            dropAction = Qt::LinkAction;
        }
        else if (event->modifiers() & Qt::ShiftModifier)
        {
            dropAction = Qt::MoveAction;
        }
        else if (event->modifiers() & Qt::ControlModifier)
        {
            dropAction = Qt::CopyAction;
        }
        else
        {
            KMenu *menu = new KMenu(qobject_cast<QWidget*>(this));
            QAction *moveAction = menu->addAction(KIcon("go-jump"), i18nc("@action:inmenu", "&Move Here\t<shortcut>%1</shortcut>", QKeySequence(Qt::ShiftModifier).toString()));
            QAction *copyAction = menu->addAction(KIcon("edit-copy"), i18nc("@action:inmenu", "&Copy Here\t<shortcut>%1</shortcut>", QKeySequence(Qt::ControlModifier).toString()));
            QAction *linkAction = menu->addAction(KIcon("insert-link"), i18nc("@action:inmenu", "&Link Here\t<shortcut>%1</shortcut>", QKeySequence(Qt::ControlModifier + Qt::ShiftModifier).toString()));

            menu->addSeparator();
            menu->addAction(KIcon("process-stop"), i18nc("@action:inmenu", "Cancel"));

            QAction *activatedAction = menu->exec(parent->popupPosition(menu->sizeHint()));

            delete menu;

            if (activatedAction == moveAction)
            {
                dropAction = Qt::MoveAction;
            }
            else if (activatedAction == copyAction)
            {
                dropAction = Qt::CopyAction;
            }
            else if (activatedAction == linkAction)
            {
                dropAction = Qt::LinkAction;
            }
            else
            {
                return;
            }
        }

        switch (dropAction)
        {
            case Qt::MoveAction:
                KIO::move(urls, m_realUrl);
            break;
            case Qt::CopyAction:
                KIO::copy(urls, m_realUrl);
            break;
            case Qt::LinkAction:
                KIO::link(urls, m_realUrl);
            break;
            default:
                return;
        }
    }
    else if (m_url.isLocalFile() && (m_mimetype->is("application/x-executable") || m_mimetype->is("application/x-shellscript") || KDesktopFile::isDesktopFile(m_url.toLocalFile())))
    {
        QString arguments;
        QString command;

        for (int i = 0; i < urls.count(); ++i)
        {
            arguments += ' ' + KShell::quoteArg(urls[i].isLocalFile()?urls[i].path():urls[i].prettyUrl());
        }

        if (KDesktopFile::isDesktopFile(m_url.toLocalFile()))
        {
            KDesktopFile desktopFile(m_url.toLocalFile());
            KConfigGroup config = desktopFile.desktopGroup();

            command = config.readPathEntry("Exec", QString());

            if (command.isEmpty())
            {
                command = KShell::quoteArg(m_url.path());
            }
        }
        else
        {
            command = KShell::quoteArg(m_url.path());
        }

        KRun::runCommand(command + ' ' + arguments, qobject_cast<QWidget*>(this));
    }
}

void FancyTasksLauncher::setUrl(KUrl url)
{
    m_realUrl = m_url = url;

    m_mimetype = KMimeType::findByUrl(m_url);

    if (m_url.isLocalFile() && KDesktopFile::isDesktopFile(m_url.toLocalFile()))
    {
        KDesktopFile desktopFile(m_url.toLocalFile());

        m_title = (desktopFile.readName().isEmpty()?m_url.fileName():desktopFile.readName());

        m_description = (desktopFile.readGenericName().isEmpty()?(desktopFile.readComment().isEmpty()?m_url.path():desktopFile.readComment()):desktopFile.readGenericName());

        m_icon = KIcon(desktopFile.readIcon());

        KConfigGroup config = desktopFile.desktopGroup();

        QString command = config.readPathEntry("Exec", QString());

        if (command.isEmpty())
        {
            QString filePath = desktopFile.readUrl();

            if (filePath.isEmpty())
            {
                filePath = desktopFile.readPath();
            }

            m_realUrl = KUrl(filePath);

            m_mimetype = KMimeType::findByUrl(KUrl(m_realUrl));
        }
    }
    else
    {
        m_title = m_url.fileName();

        m_description = m_url.path();

        m_icon = KIcon(KMimeType::iconNameForUrl(url));

        if (m_title.isEmpty())
        {
            if (m_url.isLocalFile())
            {
                m_title = m_url.directory();
            }
            else
            {
                m_title = m_url.protocol();
            }
        }
    }

    if (!KDesktopFile::isDesktopFile(m_url.toLocalFile()))
    {
        KFileItemList items;
        items.append(KFileItem(m_url, m_mimetype->name(), KFileItem::Unknown));

        qreal size = ((m_applet->iconSize() > 200)?m_applet->iconSize():200);

        KIO::PreviewJob *job = KIO::filePreview(items, size, size, 0, 0, true, true, new QStringList());

        connect(job, SIGNAL(gotPreview(const KFileItem&, const QPixmap&)), this, SLOT(setThumbnail(const KFileItem&, const QPixmap&)));
    }

    m_executable = (m_url.isLocalFile() && (m_mimetype->is("application/x-executable") || m_mimetype->is("application/x-shellscript") || KDesktopFile::isDesktopFile(m_url.toLocalFile())));

    emit changed();
}

void FancyTasksLauncher::setThumbnail(const KFileItem item, const QPixmap thumbnail)
{
    Q_UNUSED(item)

    m_thumbnail = thumbnail;

    emit changed();
}

void FancyTasksLauncher::setBrowseMenu()
{
    KMenu *menu = qobject_cast<KMenu*>(sender());

    if (menu->actions().count() > 2)
    {
        return;
    }

    QString url = menu->actions()[0]->data().toString();
    QDir dir(url);
    const QStringList entries = dir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot, QDir::LocaleAware | QDir::DirsFirst);

    foreach (const QString& entry, entries)
    {
        QString path = url;

        if (!path.endsWith('/'))
        {
            path.append('/');
        }

        path += entry;

        QString iconName = KMimeType::iconNameForUrl(KUrl(path));

        if (!QFileInfo(path).isFile())
        {
            KMenu *subMenu = new KMenu;

            QAction *action = subMenu->addAction(KIcon("document-open"), i18n("Open"));
            action->setData(path);

            subMenu->addSeparator();

            action = menu->addAction(KIcon(iconName), entry);
            action->setMenu(subMenu);

            connect(subMenu, SIGNAL(aboutToShow()), this, SLOT(setBrowseMenu()));
        }
        else
        {
            QAction *action = menu->addAction(KIcon(iconName), entry);
            action->setData(path);
        }
    }
}

void FancyTasksLauncher::openUrl(QAction *action)
{
    if (!action->data().isNull())
    {
       new KRun(KUrl(action->data().toString()), qobject_cast<QWidget*>(this));
    }
}

void FancyTasksLauncher::activate()
{
    new KRun(m_url, qobject_cast<QWidget*>(this));
}

void FancyTasksLauncher::remove()
{
    urlUpdated(m_url, KUrl());
}

void FancyTasksLauncher::showPropertiesDialog()
{
    m_propertiesDialog = new KPropertiesDialog(m_url, qobject_cast<QWidget*>(this));

    connect(m_propertiesDialog, SIGNAL(applied()), this, SLOT(propertiesDialogAccepted()));
    connect(m_propertiesDialog, SIGNAL(propertiesClosed()), this, SLOT(propertiesDialogClosed()));

    m_propertiesDialog->setWindowTitle(i18n("%1 Settings", m_title));

    m_propertiesDialog->show();
}

void FancyTasksLauncher::propertiesDialogAccepted()
{
    emit urlUpdated(m_url, m_propertiesDialog->kurl());

    m_url = m_propertiesDialog->kurl();

    setUrl(m_url);
}

void FancyTasksLauncher::propertiesDialogClosed()
{
    m_propertiesDialog = NULL;
}

void FancyTasksLauncher::removeItem()
{
    --m_itemsAmount;

    if (m_itemsAmount < 0)
    {
        m_itemsAmount = 0;
    }

    if (!m_itemsAmount)
    {
        int index = 1;

        for (int i = 0; i < m_applet->launchers().count(); ++i)
        {
            if (m_applet->launchers().at(i) == this)
            {
                break;
            }

            if (!m_applet->launchers().at(i)->itemsAmount())
            {
                ++index;
            }
        }

        m_applet->addLauncher(this, index);
    }
}

KMenu* FancyTasksLauncher::contextMenu()
{
    KMenu *menu = new KMenu(qobject_cast<QWidget*>(this));
    QAction *openAction = menu->addAction(KIcon("system-run"), i18n("Run"));

    connect(openAction, SIGNAL(triggered()), this, SLOT(activate()));

    if (m_mimetype->is("inode/directory"))
    {
        KMenu *browseMenu = new KMenu(qobject_cast<QWidget*>(this));

        QAction *action = browseMenu->addAction(KIcon("document-open"), i18n("Open"));
        action->setData(m_url.path());

        browseMenu->addSeparator();

        action = menu->addAction(KIcon("document-preview"), i18n("Browse"));
        action->setMenu(browseMenu);

        connect(browseMenu, SIGNAL(aboutToShow()), this, SLOT(setBrowseMenu()));
        connect(browseMenu, SIGNAL(triggered(QAction*)), this, SLOT(openUrl(QAction*)));
    }

    QAction *editAction = menu->addAction(KIcon("document-edit"), i18n("Edit"));

    connect(editAction, SIGNAL(triggered()), this, SLOT(showPropertiesDialog()));

    if (m_applet->immutability() == Plasma::Mutable)
    {
        menu->addSeparator();

        QAction *removeAction = menu->addAction(KIcon("list-remove"), i18n("Remove"));

        connect(removeAction, SIGNAL(triggered()), this, SLOT(remove()));
    }

    menu->setTitle(i18n("Launcher"));

    return menu;
}

void FancyTasksLauncher::addItem(FancyTasksItem *item)
{
    if (!m_itemsAmount)
    {
        m_applet->removeLauncherIcon(this);
    }

    if (!m_items.contains(item))
    {
        ++m_itemsAmount;

        m_items.append(item);

        connect(item, SIGNAL(destroyed()), this, SLOT(removeItem()));
    }
}

int FancyTasksLauncher::itemsAmount()
{
    return m_itemsAmount;
}

KIcon FancyTasksLauncher::icon()
{
    return m_icon;
}

KUrl FancyTasksLauncher::url()
{
    return m_url;
}

QPixmap FancyTasksLauncher::thumbnail()
{
    return m_thumbnail;
}

QString FancyTasksLauncher::title()
{
    return m_title;
}

QString FancyTasksLauncher::description()
{
    return m_description;
}

bool FancyTasksLauncher::isExecutable()
{
    return m_executable;
}
