/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "FancyTasksPreview.h"
#include "FancyTasksLauncher.h"

#include <QPainter>

#include <Plasma/Theme>

#include <KWindowSystem>

FancyTasksPreview::FancyTasksPreview(QPixmap pixmap, QPixmap icon, QString text, int offset, AbstractItemPtr abstractItem, QWidget *parent) : QWidget(parent)
{
    m_task = static_cast<TaskManager::TaskItem*>(abstractItem);

    m_group = NULL;

    m_launcher = NULL;

    init(pixmap, icon, text, offset);
}

FancyTasksPreview::FancyTasksPreview(QPixmap pixmap, QPixmap icon, QString text, int offset, FancyTasksLauncher *launcher, QWidget *parent) : QWidget(parent)
{
    m_task = NULL;

    m_group = NULL;

    m_launcher = launcher;

    init(pixmap, icon, text, offset);
}

FancyTasksPreview::FancyTasksPreview(QPixmap pixmap, QPixmap icon, QString text, AbstractItemPtr abstractItem, QWidget *parent) : QWidget(parent)
{
    m_task = NULL;

    m_group = static_cast<TaskManager::TaskGroup*>(abstractItem);

    m_launcher = NULL;

    init(pixmap, icon, text, 0);
}

void FancyTasksPreview::init(QPixmap pixmap, QPixmap icon, QString text, int offset)
{
    setAcceptDrops(true);

    m_pixmap = pixmap;

    m_icon = icon;

    m_offset = offset;

    m_opacity = 0.5;

    m_document = new QTextDocument(this);
    m_document->setHtml(text);
    m_document->setTextWidth((m_pixmap.width() < 100)?100:pixmap.width());
    m_document->adjustSize();

    if (!m_pixmap.isNull())
    {
        m_background = new Plasma::FrameSvg(this);
        m_background->setImagePath("widgets/frame");
        m_background->setElementPrefix("raised");

        qreal left, top, right, bottom;

        m_background->getMargins(left, top, right, bottom);

        m_previewSize = QSize((m_pixmap.width() + left + right), (m_pixmap.height() + top + bottom));

        m_background->resizeFrame(m_previewSize);

        connect(m_background, SIGNAL(repaintNeeded()), this, SLOT(updateTheme()));
    }
    else
    {
        m_previewSize = QSize(0, 0);
    }

    setFixedSize(QSize((((m_document->size().width() + m_icon.width()) > m_previewSize.width())?(m_document->size().width() + m_icon.width()):m_previewSize.width()), (m_offset + m_previewSize.height() + ((m_document->size().height() > m_icon.height())?m_document->size().height():m_icon.height()))));

    updateTheme();
}

void FancyTasksPreview::paintEvent(QPaintEvent *event)
{
    Q_UNUSED(event)

    QPainter painter(this);
    painter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    if (!m_pixmap.isNull())
    {
        painter.setOpacity(m_opacity);

        qreal left, top, right, bottom;

        m_background->getMargins(left, top, right, bottom);
        m_background->paintFrame(&painter, QPointF(((size().width() - m_previewSize.width()) / 2), m_offset));

        painter.setOpacity(1);
        painter.drawPixmap((((size().width() - m_previewSize.width()) / 2) + left), (top + m_offset), m_pixmap);
    }

    painter.drawPixmap(0, (m_previewSize.height() + m_offset + ((size().height() - m_offset - m_previewSize.height() - m_icon.height()) / 2)), m_icon);

    QPixmap description(m_document->size().toSize());
    description.fill(Qt::transparent);

    QPainter descriptionPainter(&description);
    descriptionPainter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing | QPainter::TextAntialiasing);

    m_document->drawContents(&descriptionPainter);

    descriptionPainter.end();

    painter.drawPixmap(m_icon.width(), (m_previewSize.height() + m_offset), description);
}

void FancyTasksPreview::enterEvent(QEvent *event)
{
    Q_UNUSED(event)

    m_opacity = 1;

    update();
}

void FancyTasksPreview::leaveEvent(QEvent *event)
{
    Q_UNUSED(event)

    m_opacity = 0.5;

    update();
}

void FancyTasksPreview::mousePressEvent(QMouseEvent *event)
{
    event->ignore();

    m_opacity = 0.5;

    if (!m_task && !m_launcher)
    {
        return;
    }

    if (event->button() == Qt::LeftButton)
    {
        if (m_task)
        {
            m_task->task()->activateRaiseOrIconify();
        }
        else
        {
            m_launcher->activate();
        }
    }
    else if (event->button() == Qt::MidButton && m_task)
    {
        m_task->task()->close();
    }
    else if (event->button() == Qt::RightButton)
    {
        if (m_task || m_group)
        {
            QList<QAction*> actionList;
            TaskManager::BasicMenu *taskMenu;

            if (m_task)
            {
                taskMenu = new TaskManager::BasicMenu(this, m_task, new TaskManager::GroupManager(NULL), actionList);
            }
            else
            {
                taskMenu = new TaskManager::BasicMenu(this, m_group, new TaskManager::GroupManager(NULL), actionList);
            }

            if (m_launcher)
            {
                taskMenu->addSeparator();

                taskMenu->addMenu(m_launcher->contextMenu());
            }

            taskMenu->exec(mapToGlobal(event->pos()));

            delete taskMenu;
        }
        else
        {
            KMenu *launcherMenu = m_launcher->contextMenu();

            launcherMenu->exec(mapToGlobal(event->pos()));

            delete launcherMenu;
        }
    }
}

void FancyTasksPreview::dragEnterEvent(QDragEnterEvent *event)
{
    if (m_task)
    {
        m_task->task()->activateRaiseOrIconify();
    }

    event->ignore();
}

void FancyTasksPreview::updateTheme()
{
    if (!m_pixmap.isNull())
    {
        m_background->clearCache();
    }

    QPalette plasmaPalette = QPalette();
    plasmaPalette.setColor(QPalette::Window, Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor));
    plasmaPalette.setColor(QPalette::WindowText, Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor));

    m_document->setDefaultStyleSheet(QString("p { color: %1; }").arg(Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor).name()));
    m_document->setHtml(m_document->toHtml());

    update();
}

TaskManager::TaskItem* FancyTasksPreview::task()
{
    return m_task;
}

Plasma::FrameSvg* FancyTasksPreview::background()
{
    return m_background;
}

QSize FancyTasksPreview::previewSize()
{
    return m_previewSize;
}

int FancyTasksPreview::offset()
{
    return m_offset;
}
