/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "FancyTasksSpacer.h"
#include "FancyTasksApplet.h"
#include "FancyTasksIcon.h"
#include "FancyTasksItem.h"

#include <QPainter>
#include <QTimer>

#include <KUrl>

#include <Plasma/Theme>
#include <Plasma/PaintUtils>

FancyTasksSpacer::FancyTasksSpacer(FancyTasksApplet *applet)
{
    setAcceptDrops(true);

    m_applet = applet;

    m_index = 0;

    m_canHide = true;

    m_visible = false;

    setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
}

void FancyTasksSpacer::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option)
    Q_UNUSED(widget)

    if (!m_visible)
    {
        return;
    }

    QPainterPath path = Plasma::PaintUtils::roundedRectangle(contentsRect().adjusted(1, 1, -2, -2), 4);
    QColor color = Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor);
    color.setAlphaF(0.3);

    painter->setRenderHint(QPainter::Antialiasing);
    painter->fillPath(path, color);
}

void FancyTasksSpacer::dragEnterEvent(QGraphicsSceneDragDropEvent *event)
{
    Q_UNUSED(event)

    m_canHide = false;
}

void FancyTasksSpacer::dragLeaveEvent(QGraphicsSceneDragDropEvent *event)
{
    Q_UNUSED(event)

    m_canHide = true;

    QTimer::singleShot(1000, this, SLOT(hide()));
}

void FancyTasksSpacer::dropEvent(QGraphicsSceneDragDropEvent *event)
{
    m_canHide = true;

    m_lastShow = QDateTime();

    hide();

    if (event->mimeData()->hasFormat("FancyTasksIcon"))
    {
        QByteArray data = event->mimeData()->data("FancyTasksIcon");

        FancyTasksIcon *droppedIcon;

        memcpy(&droppedIcon, data, sizeof(FancyTasksIcon*));

        if (droppedIcon->item()->itemType() == FancyTasksItem::Task || droppedIcon->item()->itemType() == FancyTasksItem::Group)
        {
            m_index -= m_applet->launcherIcons().count();

            if (m_index < 0)
            {
                m_index = 0;
            }

            TaskManager::AbstractGroupableItem *temp = static_cast<TaskManager::AbstractGroupableItem*>(droppedIcon->item()->abstractItem());

            m_applet->groupManager()->manualSortingRequest(temp, m_index);

            return;
        }

        if (m_applet->immutability() != Plasma::Mutable)
        {
            return;
        }

        if (m_index >= m_applet->launcherIcons().count())
        {
            m_index = (m_applet->launcherIcons().count()?(m_applet->launcherIcons().count() - 1):0);
        }

        m_applet->layout()->removeItem(droppedIcon);
        m_applet->layout()->insertItem(m_index, droppedIcon);

        m_applet->saveConfig();

        return;
    }

    if (!KUrl::List::canDecode(event->mimeData()) || m_applet->immutability() != Plasma::Mutable)
    {
        event->ignore();

        return;
    }

    KUrl::List droppedUrls = KUrl::List::fromMimeData(event->mimeData());

    for (int i = 0; i < droppedUrls.count(); ++i)
    {
        m_applet->updateUrls(KUrl(), droppedUrls[i]);
    }

    m_applet->saveConfig();
}

void FancyTasksSpacer::setSize(qreal size)
{
    m_size = size;
}

void FancyTasksSpacer::setIndex(int index)
{
    m_index = index;
}

void FancyTasksSpacer::show()
{
    m_lastShow = QDateTime::currentDateTime();

    if (m_visible)
    {
        return;
    }

    m_visible = true;

    m_applet->layout()->removeItem(this);
    m_applet->layout()->insertItem(m_index, this);

    if (m_applet->formFactor() == Plasma::Vertical)
    {
        setPreferredSize(m_size, (m_size / 2));
    }
    else
    {
        setPreferredSize((m_size / 2), m_size);
    }

    m_applet->updateSize();
}

void FancyTasksSpacer::hide()
{
    if (m_lastShow.isValid() && m_lastShow.secsTo(QDateTime::currentDateTime()) < 1)
    {
        QTimer::singleShot(200, this, SLOT(hide()));

        return;
    }

    if (m_canHide)
    {
        m_visible = false;

        m_applet->layout()->removeItem(this);
        m_applet->layout()->insertItem(0, this);

        setPreferredSize(0, 0);
    }
}

int FancyTasksSpacer::index()
{
    return (m_visible?m_index:0);
}

bool FancyTasksSpacer::isVisible()
{
    return m_visible;
}
