/***********************************************************************************
* Fancy Tasks: Plasmoid for fancy representing your tasks and launchers.
* Copyright (C) 2009 Michal Dutkiewicz aka Emdek <emdeck@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

#include "FancyTasksToolTip.h"
#include "FancyTasksPreview.h"
#include "FancyTasksItem.h"
#include "FancyTasksLauncher.h"

#include <Plasma/Theme>

#include <QTimer>

#ifdef Q_WS_X11
#include <QX11Info>

#include <X11/Xlib.h>
#include <fixx11h.h>
#endif

FancyTasksToolTip::FancyTasksToolTip(QPixmap pixmap, QPixmap icon, QString description, AbstractItemPtr abstractItem, int size, QWidget *parent) : QWidget(parent)
{
    setWindowFlags(Qt::ToolTip);

    m_size = size;

    init();

#ifdef Q_WS_X11
    if (KWindowSystem::compositingActive())
    {
        int number;
        Display *display = QX11Info::display();
        Atom atom = XInternAtom(display, "_KDE_WINDOW_PREVIEW", false);
        Atom *list = XListProperties(display, DefaultRootWindow(display), &number);

        if (list != NULL)
        {
            m_previewsAvailable = (qFind(list, (list + number), atom) != (list + number));

            XFree(list);
        }
    }
#endif

    if (!m_size)
    {
        pixmap = NULL;
    }

    if (abstractItem->isGroupItem())
    {
        TaskManager::TaskGroup *group = static_cast<TaskManager::TaskGroup*>(abstractItem);

        m_tasksLayout = new QGridLayout;

        m_mainLayout->addLayout(m_tasksLayout, (m_previewsAvailable?0:2), 0, 1, 2);

        TaskManager::ItemList tasks = group->members();
        QList<QPixmap> pixmaps;

        int maxHeight = 0;

        for (int i = 0; i < tasks.count(); ++i)
        {
            TaskManager::TaskItem *task = static_cast<TaskManager::TaskItem*>(tasks.at(i));

            if (m_previewsAvailable && m_size)
            {
                if (task->task()->window())
                {
                    QSize windowSize = KWindowSystem::windowInfo(task->task()->window(), NET::WMGeometry | NET::WMFrameExtents).frameGeometry().size();

                    if (windowSize.isValid())
                    {
                        if (windowSize.width() > m_size || windowSize.height() > m_size)
                        {
                            windowSize.scale(m_size, m_size, Qt::KeepAspectRatio);
                        }

                        pixmap = QPixmap(windowSize);
                        pixmap.fill(Qt::transparent);
                    }
                }
            }

            if (!pixmap.isNull() && pixmap.height() > maxHeight)
            {
                maxHeight = pixmap.height();
            }

            pixmaps.append(pixmap);
        }

        for (int i = 0; i < tasks.count(); ++i)
        {
            TaskManager::TaskItem *task = static_cast<TaskManager::TaskItem*>(tasks.at(i));

            FancyTasksPreview *preview = new FancyTasksPreview(pixmaps.at(i), tasks.at(i)->icon().pixmap(16, 16), task->task()->visibleName(), (maxHeight - pixmaps.at(i).height()), task, this);

            m_previews.append(preview);

            m_tasksLayout->addWidget(preview, (m_previewsAvailable?0:i), (m_previewsAvailable?i:0), Qt::AlignTop);
        }

        FancyTasksPreview *preview = new FancyTasksPreview(QPixmap(), icon, description, abstractItem, this);

        m_mainLayout->addWidget(preview, 1, 0, Qt::AlignLeft);
    }
    else
    {
        TaskManager::TaskItem *task = static_cast<TaskManager::TaskItem*>(abstractItem);

        if (m_previews.count())
        {
            m_previews.at(0)->deleteLater();
            m_previews.removeAt(0);
        }

        if (m_previewsAvailable && m_size && task->task()->window())
        {
            QSize windowSize = KWindowSystem::windowInfo(task->task()->window(), NET::WMGeometry | NET::WMFrameExtents).frameGeometry().size();

            if (windowSize.isValid())
            {
                if (windowSize.width() > m_size || windowSize.height() > m_size)
                {
                    windowSize.scale(m_size, m_size, Qt::KeepAspectRatio);
                }

                pixmap = QPixmap(windowSize);
                pixmap.fill(Qt::transparent);
            }
        }

        FancyTasksPreview *preview = new FancyTasksPreview(pixmap, icon, description, 0, abstractItem, this);

        m_previews.append(preview);

        m_mainLayout->addWidget(preview, 0, 0, Qt::AlignCenter);
    }

    adjustSize();
}

FancyTasksToolTip::FancyTasksToolTip(QPixmap pixmap, QPixmap icon, QString description, FancyTasksLauncher* launcher, int size, QWidget *parent) : QWidget(parent)
{
    setWindowFlags(Qt::ToolTip);

    m_size = size;

    init();

    if (!m_size)
    {
        pixmap = NULL;
    }

    FancyTasksPreview *preview = new FancyTasksPreview(pixmap, icon, description, 0, launcher, this);

    m_previews.append(preview);

    m_mainLayout->addWidget(preview, 0, 0, Qt::AlignCenter);

    adjustSize();
}

FancyTasksToolTip::~FancyTasksToolTip()
{
    delete m_mainLayout;
    delete m_background;
}

void FancyTasksToolTip::init()
{
    m_canHide = true;

    m_alreadyShown = false;

    m_previewsAvailable = false;

    m_background = new Plasma::FrameSvg(this);
    m_background->setImagePath("widgets/tooltip");
    m_background->setEnabledBorders(Plasma::FrameSvg::AllBorders);

    m_mainLayout = new QGridLayout(this);
    m_mainLayout->setSpacing(3);

    setLayout(m_mainLayout);

    updateTheme();

    connect(m_background, SIGNAL(repaintNeeded()), this, SLOT(updateTheme()));
}

void FancyTasksToolTip::paintEvent(QPaintEvent *event)
{
    Q_UNUSED(event)

    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing);
    painter.setClipRect(event->rect());
    painter.setCompositionMode(QPainter::CompositionMode_Source);
    painter.fillRect(rect(), Qt::transparent);

    m_background->paintFrame(&painter);
}

void FancyTasksToolTip::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);

    m_background->resizeFrame(size());

    setMask(m_background->mask());
}

void FancyTasksToolTip::enterEvent(QEvent *event)
{
    Q_UNUSED(event)

    m_canHide = false;
}

void FancyTasksToolTip::leaveEvent(QEvent *event)
{
    Q_UNUSED(event)

    m_canHide = true;

    QTimer::singleShot(500, this, SLOT(hide()));
}

void FancyTasksToolTip::mousePressEvent(QMouseEvent *event)
{
    m_canHide = true;

    hide();

    event->accept();
}

void FancyTasksToolTip::show()
{
#ifdef Q_WS_X11
    if (m_alreadyShown || (m_previews.count() == 1 && !m_previews[0]->task()) || !m_size || !m_previewsAvailable)
    {
        QWidget::show();

        return;
    }

    m_alreadyShown = true;

    for (int i = 0; i < m_previews.count(); ++i)
    {
        m_previews.at(i)->show();
    }

    m_mainLayout->activate();

    Display *display = QX11Info::display();
    long unsigned int atom = XInternAtom(display, "_KDE_WINDOW_PREVIEW", false);
    long data[(m_previews.count() * 6) + 1];
    data[0] = m_previews.count();

    for (int i = 0; i < m_previews.count(); ++i)
    {
        int index = (i * 6);

        qreal left, top, right, bottom;

        if (!m_previews.at(i)->background())
        {
            continue;
        }

        m_previews.at(i)->background()->getMargins(left, top, right, bottom);

        QPoint position = m_previews.at(i)->geometry().topLeft();
        position.setX(position.x() + ((m_previews.at(i)->size().width() - m_previews.at(i)->previewSize().width()) / 2));
        position.setY(position.y() + m_previews.at(i)->offset());

        QRect thumbnailRect = QRect(position, m_previews.at(i)->previewSize()).adjusted(left, top, -right, -bottom);

        data[index + 1] = 5;
        data[index + 2] = m_previews.at(i)->task()->task()->window();
        data[index + 3] = thumbnailRect.left();
        data[index + 4] = thumbnailRect.top();
        data[index + 5] = thumbnailRect.width();
        data[index + 6] = thumbnailRect.height();
    }

    XChangeProperty(display, winId(), atom, atom, 32, PropModeReplace, reinterpret_cast<unsigned char*>(data), (sizeof(data) / sizeof(data[0])));
#endif

    QWidget::show();
}

void FancyTasksToolTip::hide(bool force)
{
    if (m_canHide || force)
    {
        m_canHide = true;

        QWidget::hide();
    }
}

void FancyTasksToolTip::updateTheme()
{
    m_background->clearCache();

    m_mainLayout->setContentsMargins(m_background->marginSize(Plasma::LeftMargin), m_background->marginSize(Plasma::TopMargin), m_background->marginSize(Plasma::RightMargin), m_background->marginSize(Plasma::BottomMargin));

    QPalette plasmaPalette = QPalette();
    plasmaPalette.setColor(QPalette::Window, Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor));
    plasmaPalette.setColor(QPalette::WindowText, Plasma::Theme::defaultTheme()->color(Plasma::Theme::TextColor));

    setAutoFillBackground(true);

    setPalette(plasmaPalette);

    update();
}

void FancyTasksToolTip::setCanHide(bool canHide)
{
    m_canHide = canHide;
}
